#------------------------------------------------------------------------------
# Copyright (c) 2020, 2023, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
2400 - Module for testing pools
"""

import threading
import unittest

import oracledb
import test_env

class TestCase(test_env.BaseTestCase):
    require_connection = False

    def __connect_and_drop(self):
        with self.pool.acquire() as connection:
            cursor = connection.cursor()
            cursor.execute("select count(*) from TestNumbers")
            count, = cursor.fetchone()
            self.assertEqual(count, 10)

    def __connect_and_generate_error(self):
        with self.pool.acquire() as connection:
            cursor = connection.cursor()
            self.assertRaisesRegex(oracledb.DatabaseError,"^ORA-01476:",
                                   cursor.execute, "select 1 / 0 from dual")

    def __callable_session_callback(self, conn, requested_tag):
        self.session_called = True

        supported_formats = {
            "SIMPLE" : "'YYYY-MM-DD HH24:MI'",
            "FULL" : "'YYYY-MM-DD HH24:MI:SS'"
        }

        supported_time_zones = {
            "UTC" : "'UTC'",
            "MST" : "'-07:00'"
        }

        supported_keys = {
            "NLS_DATE_FORMAT" : supported_formats,
            "TIME_ZONE" : supported_time_zones
        }
        if requested_tag is not None:
            state_parts = []
            for directive in requested_tag.split(";"):
                parts = directive.split("=")
                if len(parts) != 2:
                    raise ValueError("Tag must contain key=value pairs")
                key, value = parts
                value_dict = supported_keys.get(key)
                if value_dict is None:
                    raise ValueError("Tag only supports keys: %s" % \
                                     (", ".join(supported_keys)))
                actual_value = value_dict.get(value)
                if actual_value is None:
                    raise ValueError("Key %s only supports values: %s" % \
                                     (key, ", ".join(value_dict)))
                state_parts.append("%s = %s" % (key, actual_value))
            sql = "alter session set %s" % " ".join(state_parts)
            cursor = conn.cursor()
            cursor.execute(sql)
        conn.tag = requested_tag

    def __perform_reconfigure_test(self, parameter_name, parameter_value,
                                   min=3, max=30, increment=4, timeout=5,
                                   wait_timeout=5000, stmtcachesize=25,
                                   max_lifetime_session=1000,
                                   max_sessions_per_shard=3, ping_interval=30,
                                   getmode=oracledb.POOL_GETMODE_WAIT):
        creation_args = dict(min=min, max=max, increment=increment,
                             timeout=timeout, wait_timeout=wait_timeout,
                             stmtcachesize=stmtcachesize,
                             max_lifetime_session=max_lifetime_session,
                             max_sessions_per_shard=max_sessions_per_shard,
                             ping_interval=ping_interval, getmode=getmode)
        reconfigure_args = {}
        reconfigure_args[parameter_name] = parameter_value

        pool = test_env.get_pool(**creation_args)
        connection = pool.acquire()
        pool.reconfigure(**reconfigure_args)
        actual_args = dict(min=pool.min, max=pool.max,
                           increment=pool.increment, timeout=pool.timeout,
                           wait_timeout=pool.wait_timeout,
                           stmtcachesize=pool.stmtcachesize,
                           max_lifetime_session=pool.max_lifetime_session,
                           max_sessions_per_shard=pool.max_sessions_per_shard,
                           ping_interval=pool.ping_interval,
                           getmode=pool.getmode)
        expected_args = creation_args.copy()
        expected_args.update(reconfigure_args)
        self.assertEqual(actual_args, expected_args)

    def __verify_connection(self, connection, expected_user,
                            expected_proxy_user=None):
        cursor = connection.cursor()
        cursor.execute("""
                select
                    sys_context('userenv', 'session_user'),
                    sys_context('userenv', 'proxy_user')
                from dual""")
        actual_user, actual_proxy_user = cursor.fetchone()
        self.assertEqual(actual_user, expected_user.upper())
        self.assertEqual(actual_proxy_user,
                         expected_proxy_user and expected_proxy_user.upper())

    def test_2400_pool(self):
        "2400 - test that the pool is created and has the right attributes"
        pool = test_env.get_pool(min=2, max=8, increment=3,
                                 getmode=oracledb.POOL_GETMODE_WAIT)
        self.assertEqual(pool.username, test_env.get_main_user(),
                         "user name differs")
        self.assertEqual(pool.dsn, test_env.get_connect_string(),
                         "dsn differs")
        self.assertEqual(pool.max, 8, "max differs")
        self.assertEqual(pool.min, 2, "min differs")
        self.assertEqual(pool.increment, 3, "increment differs")
        self.assertEqual(pool.busy, 0, "busy not 0 at start")
        connection_1 = pool.acquire()
        self.assertEqual(pool.busy, 1, "busy not 1 after acquire")
        connection_2 = oracledb.connect(pool=pool)
        self.assertEqual(pool.busy, 2, "busy not 2 after acquire")
        self.assertEqual(pool.opened, 2, "opened differs")
        connection_3 = pool.acquire()
        self.assertEqual(pool.busy, 3, "busy not 3 after acquire")
        pool.release(connection_3)
        self.assertEqual(pool.busy, 2, "busy not 2 after release")
        pool.release(connection_1)
        connection_2.close()
        self.assertEqual(pool.busy, 0, "busy not 0 after release")
        pool.getmode = oracledb.POOL_GETMODE_NOWAIT
        self.assertEqual(pool.getmode, oracledb.POOL_GETMODE_NOWAIT)
        if test_env.get_client_version() >= (12, 2):
            pool.getmode = oracledb.POOL_GETMODE_TIMEDWAIT
            self.assertEqual(pool.getmode, oracledb.POOL_GETMODE_TIMEDWAIT)
        pool.stmtcachesize = 50
        self.assertEqual(pool.stmtcachesize, 50)
        pool.timeout = 10
        self.assertEqual(pool.timeout, 10)
        if test_env.get_client_version() >= (12, 1):
            pool.max_lifetime_session = 10
            self.assertEqual(pool.max_lifetime_session, 10)

    @unittest.skipIf(test_env.get_is_thin(),
                     "thin mode doesn't support proxy users yet")
    def test_2401_proxy_auth(self):
        "2401 - test that proxy authentication is possible"
        pool = test_env.get_pool(min=2, max=8, increment=3,
                                 getmode=oracledb.POOL_GETMODE_WAIT)
        self.assertEqual(pool.homogeneous, True,
                         "homogeneous should be True by default")
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPI-1012:",
                               pool.acquire, user="missing_proxyuser")
        pool = test_env.get_pool(min=2, max=8, increment=3,
                                 getmode=oracledb.POOL_GETMODE_WAIT,
                                 homogeneous=False)
        msg = "homogeneous should be False after setting it in the constructor"
        self.assertEqual(pool.homogeneous, False, msg)
        connection = pool.acquire(user=test_env.get_proxy_user())
        cursor = connection.cursor()
        cursor.execute('select user from dual')
        result, = cursor.fetchone()
        self.assertEqual(result, test_env.get_proxy_user().upper())
        connection.close()

    def test_2403_rollback_on_release(self):
        "2403 - connection rolls back before released back to the pool"
        pool = test_env.get_pool(getmode=oracledb.POOL_GETMODE_WAIT)
        connection = pool.acquire()
        cursor = connection.cursor()
        cursor.execute("truncate table TestTempTable")
        cursor.execute("insert into TestTempTable (IntCol) values (1)")
        cursor.close()
        pool.release(connection)
        pool = test_env.get_pool(getmode=oracledb.POOL_GETMODE_WAIT)
        connection = pool.acquire()
        cursor = connection.cursor()
        cursor.execute("select count(*) from TestTempTable")
        count, = cursor.fetchone()
        self.assertEqual(count, 0)
        connection.close()

    def test_2404_threading(self):
        "2404 - test session pool with multiple threads"
        self.pool = test_env.get_pool(min=5, max=20, increment=2,
                                      getmode=oracledb.POOL_GETMODE_WAIT)
        threads = []
        for i in range(20):
            thread = threading.Thread(None, self.__connect_and_drop)
            threads.append(thread)
            thread.start()
        for thread in threads:
            thread.join()

    def test_2405_threading_with_errors(self):
        "2405 - test session pool with multiple threads (with errors)"
        self.pool = test_env.get_pool(min=5, max=20, increment=2,
                                      getmode=oracledb.POOL_GETMODE_WAIT)
        threads = []
        for i in range(20):
            thread = threading.Thread(None, self.__connect_and_generate_error)
            threads.append(thread)
            thread.start()
        for thread in threads:
            thread.join()

    def test_2406_purity(self):
        "2406 - test session pool with various types of purity"
        pool = test_env.get_pool(min=1, max=8, increment=1,
                                 getmode=oracledb.POOL_GETMODE_WAIT)

        # get connection and set the action
        action = "TEST_ACTION"
        connection = pool.acquire()
        connection.action = action
        cursor = connection.cursor()
        cursor.execute("select 1 from dual")
        cursor.close()
        pool.release(connection)
        self.assertEqual(pool.opened, 1, "opened (1)")

        # verify that the connection still has the action set on it
        connection = pool.acquire()
        cursor = connection.cursor()
        cursor.execute("select sys_context('userenv', 'action') from dual")
        result, = cursor.fetchone()
        self.assertEqual(result, action)
        cursor.close()
        pool.release(connection)
        self.assertEqual(pool.opened, 1, "opened (2)")

        # get a new connection with new purity (should not have state)
        connection = pool.acquire(purity=oracledb.ATTR_PURITY_NEW)
        cursor = connection.cursor()
        cursor.execute("select sys_context('userenv', 'action') from dual")
        result, = cursor.fetchone()
        self.assertEqual(result, None)
        cursor.close()
        pool.release(connection)

    @unittest.skipIf(test_env.get_is_thin(),
                     "thin mode doesn't support proxy users yet")
    def test_2407_heterogeneous(self):
        "2407 - test heterogeneous pool with user and password specified"
        pool = test_env.get_pool(min=2, max=8, increment=3, homogeneous=False,
                                 getmode=oracledb.POOL_GETMODE_WAIT)
        self.assertEqual(pool.homogeneous, 0)
        conn = pool.acquire()
        self.__verify_connection(pool.acquire(), test_env.get_main_user())
        conn.close()
        conn = pool.acquire(test_env.get_main_user(),
                            test_env.get_main_password())
        self.__verify_connection(conn, test_env.get_main_user())
        conn.close()
        conn = pool.acquire(test_env.get_proxy_user(),
                            test_env.get_proxy_password())
        self.__verify_connection(conn, test_env.get_proxy_user())
        conn.close()
        user_str = "%s[%s]" % \
                   (test_env.get_main_user(), test_env.get_proxy_user())
        conn = pool.acquire(user_str, test_env.get_main_password())
        self.__verify_connection(conn, test_env.get_proxy_user(),
                                 test_env.get_main_user())
        conn.close()

    @unittest.skipIf(test_env.get_is_thin(),
                     "thin mode doesn't support proxy users yet")
    def test_2408_heterogenous_without_user(self):
        "2408 - test heterogeneous pool without user and password specified"
        pool = test_env.get_pool(user="", password="", min=2, max=8,
                                 increment=3,
                                 getmode=oracledb.POOL_GETMODE_WAIT,
                                 homogeneous=False)
        conn = pool.acquire(test_env.get_main_user(),
                            test_env.get_main_password())
        self.__verify_connection(conn, test_env.get_main_user())
        conn.close()
        conn = pool.acquire(test_env.get_proxy_user(),
                            test_env.get_proxy_password())
        self.__verify_connection(conn, test_env.get_proxy_user())
        conn.close()
        user_str = "%s[%s]" % \
                   (test_env.get_main_user(), test_env.get_proxy_user())
        conn = pool.acquire(user_str, test_env.get_main_password())
        self.__verify_connection(conn, test_env.get_proxy_user(),
                                 test_env.get_main_user())

    @unittest.skipIf(test_env.get_is_thin(),
                     "thin mode doesn't support proxy users yet")
    def test_2409_heterogeneous_wrong_password(self):
        "2409 - test heterogeneous pool with wrong password specified"
        pool = test_env.get_pool(min=2, max=8, increment=3,
                                 getmode=oracledb.POOL_GETMODE_WAIT,
                                 homogeneous=False)
        self.assertRaisesRegex(oracledb.DatabaseError, "^ORA-01017:",
                               pool.acquire, test_env.get_proxy_user(),
                               "this is the wrong password")

    @unittest.skipIf(test_env.get_is_thin(),
                     "thin mode doesn't support tagging yet")
    def test_2410_tagging_session(self):
        "2410 - test tagging a session"
        pool = test_env.get_pool(min=2, max=8, increment=3,
                                 getmode=oracledb.POOL_GETMODE_NOWAIT)

        tag_mst = "TIME_ZONE=MST"
        tag_utc = "TIME_ZONE=UTC"

        conn = pool.acquire()
        self.assertEqual(conn.tag, None)
        pool.release(conn, tag=tag_mst)

        conn = pool.acquire()
        self.assertEqual(conn.tag, None)
        conn.tag = tag_utc
        conn.close()

        conn = pool.acquire(tag=tag_mst)
        self.assertEqual(conn.tag, tag_mst)
        conn.close()

        conn = pool.acquire(tag=tag_utc)
        self.assertEqual(conn.tag, tag_utc)
        conn.close()

    @unittest.skipIf(test_env.get_is_thin(),
                     "thin mode doesn't support session callbacks yet")
    def test_2411_plsql_session_callbacks(self):
        "2411 - test PL/SQL session callbacks"
        if test_env.get_client_version() < (12, 2):
            self.skipTest("PL/SQL session callbacks not supported before 12.2")
        callback = "pkg_SessionCallback.TheCallback"
        pool = test_env.get_pool(min=2, max=8, increment=3,
                                 getmode=oracledb.POOL_GETMODE_NOWAIT,
                                 session_callback=callback)
        tags = [
            "NLS_DATE_FORMAT=SIMPLE",
            "NLS_DATE_FORMAT=FULL;TIME_ZONE=UTC",
            "NLS_DATE_FORMAT=FULL;TIME_ZONE=MST"
        ]
        actual_tags = [None, None, "NLS_DATE_FORMAT=FULL;TIME_ZONE=UTC"]

        # truncate PL/SQL session callback log
        conn = pool.acquire()
        cursor = conn.cursor()
        cursor.execute("truncate table PLSQLSessionCallbacks")
        conn.close()

        # request sessions with each of the first two tags
        for tag in tags[:2]:
            conn = pool.acquire(tag=tag)
            conn.close()

        # for the last tag, use the matchanytag flag
        conn = pool.acquire(tag=tags[2], matchanytag=True)
        conn.close()

        # verify the PL/SQL session callback log is accurate
        conn = pool.acquire()
        cursor = conn.cursor()
        cursor.execute("""
                select RequestedTag, ActualTag
                from PLSQLSessionCallbacks
                order by FixupTimestamp""")
        results = cursor.fetchall()
        expected_results = list(zip(tags, actual_tags))
        self.assertEqual(results, expected_results)
        conn.close()

    @unittest.skipIf(test_env.get_is_thin(),
                     "thin mode doesn't support tagging yet")
    def test_2412_tagging_invalid_key(self):
        "2412 - testTagging with Invalid key"
        pool = test_env.get_pool(getmode=oracledb.POOL_GETMODE_NOWAIT)
        conn = pool.acquire()
        self.assertRaises(TypeError, pool.release, conn, tag=12345)
        if test_env.get_client_version() >= (12, 2):
            self.assertRaisesRegex(oracledb.DatabaseError, "^ORA-24488:",
                                   pool.release, conn, tag="INVALID_TAG")

    def test_2413_close_and_drop_connection_from_pool(self):
        "2413 - test dropping/closing a connection from the pool"
        pool = test_env.get_pool(min=1, max=8, increment=1,
                                 getmode=oracledb.POOL_GETMODE_WAIT)
        conn = pool.acquire()
        self.assertEqual(pool.busy, 1, "busy (1)")
        self.assertEqual(pool.opened, 1, "opened (1)")
        pool.drop(conn)
        self.assertEqual(pool.busy, 0, "busy (2)")
        self.assertEqual(pool.opened, 0, "opened (2)")
        conn = pool.acquire()
        self.assertEqual(pool.busy, 1, "busy (3)")
        self.assertEqual(pool.opened, 1, "opened (3)")
        conn.close()
        self.assertEqual(pool.busy, 0, "busy (4)")
        self.assertEqual(pool.opened, 1, "opened (4)")

    def test_2414_create_new_pure_connection(self):
        "2414 - test to ensure pure connections are being created correctly"
        pool = test_env.get_pool(min=1, max=2, increment=1,
                                 getmode=oracledb.POOL_GETMODE_WAIT)
        connection_1 = pool.acquire()
        connection_2 = pool.acquire()
        self.assertEqual(pool.opened, 2, "opened (1)")
        pool.release(connection_1)
        pool.release(connection_2)
        connection_3 = pool.acquire(purity=oracledb.ATTR_PURITY_NEW)
        self.assertEqual(pool.opened, 2, "opened (2)")
        pool.release(connection_3)

    @unittest.skipIf(test_env.get_is_thin(),
                     "thin mode doesn't support pool reconfigure yet")
    def test_2415_reconfigure_pool(self):
        "2415 - test to ensure reconfigure() updates pool properties"
        pool = test_env.get_pool(min=1, max=2, increment=1,
                                 getmode=oracledb.POOL_GETMODE_WAIT)
        self.assertEqual(pool.min, 1, "min (1)")
        self.assertEqual(pool.max, 2, "max (2)")
        self.assertEqual(pool.increment, 1, "increment (1)")
        self.assertEqual(pool.getmode, oracledb.POOL_GETMODE_WAIT,
                         "getmode differs")
        self.assertEqual(pool.timeout, 0, "timeout (0)")
        self.assertEqual(pool.wait_timeout, 5000, "wait_timeout (5000)")
        self.assertEqual(pool.max_lifetime_session, 0,
                         "max_lifetime_sessionmeout (0)")
        self.assertEqual(pool.max_sessions_per_shard, 0,
                         "max_sessions_per_shard (0)")
        self.assertEqual(pool.stmtcachesize, 20, "stmtcachesize (20)")
        self.assertEqual(pool.ping_interval, 60, "ping_interval (60)")

        pool.reconfigure(min=2, max=5, increment=2, timeout=30,
                         getmode=oracledb.POOL_GETMODE_TIMEDWAIT,
                         wait_timeout=3000, max_lifetime_session=20,
                         max_sessions_per_shard=2, stmtcachesize=30,
                         ping_interval=30)
        self.assertEqual(pool.min, 2, "min (2)")
        self.assertEqual(pool.max, 5, "max (5)")
        self.assertEqual(pool.increment, 2, "increment (2)")
        self.assertEqual(pool.getmode, oracledb.POOL_GETMODE_TIMEDWAIT,
                         "getmode differs")
        self.assertEqual(pool.timeout, 30, "timeout (30)")
        self.assertEqual(pool.wait_timeout, 3000, "wait_timeout (3000)")
        self.assertEqual(pool.max_lifetime_session, 20,
                         "max_lifetime_sessionmeout (20)")
        self.assertEqual(pool.max_sessions_per_shard, 2,
                         "max_sessions_per_shard (2)")
        self.assertEqual(pool.stmtcachesize, 30, "stmtcachesize (30)")
        self.assertEqual(pool.ping_interval, 30, "ping_interval (30)")

    @unittest.skipIf(test_env.get_is_thin(),
                     "thin mode doesn't support all the pool params yet")
    def test_2416_test_reconfigure_pool_with_missing_values(self):
        "2416 - test the reconfigure values are changed and rest unchanged"
        self.__perform_reconfigure_test("min", 5)
        self.__perform_reconfigure_test("max", 20)
        self.__perform_reconfigure_test("increment", 5)
        self.__perform_reconfigure_test("timeout", 10)
        self.__perform_reconfigure_test("wait_timeout", 8000)
        self.__perform_reconfigure_test("stmtcachesize", 40)
        self.__perform_reconfigure_test("max_lifetime_session", 2000)
        self.__perform_reconfigure_test("max_sessions_per_shard", 5)
        self.__perform_reconfigure_test("ping_interval", 50)
        self.__perform_reconfigure_test("getmode",
                                        oracledb.POOL_GETMODE_NOWAIT)

    @unittest.skipIf(test_env.get_is_thin(),
                     "thin mode doesn't support all the pool params yet")
    def test_2417_setting_each_pool_param(self):
        "2417 - test to see if specified parameters are set during creation"
        pool = test_env.get_pool(min=1, max=2, increment=1, timeout=10,
                                 wait_timeout=10, max_lifetime_session=20,
                                 max_sessions_per_shard=1, stmtcachesize=25,
                                 ping_interval=25,
                                 getmode=oracledb.POOL_GETMODE_WAIT)
        self.assertEqual(pool.min, 1, "min (1)")
        self.assertEqual(pool.max, 2, "max (2)")
        self.assertEqual(pool.increment, 1, "increment (1)")
        self.assertEqual(pool.getmode, oracledb.POOL_GETMODE_WAIT,
                         "getmode differs")
        self.assertEqual(pool.timeout, 10, "timeout (10)")
        self.assertEqual(pool.wait_timeout, 10, "wait_timeout (10)")
        self.assertEqual(pool.max_lifetime_session, 20,
                         "max_lifetime_sessionmeout (20)")
        self.assertEqual(pool.max_sessions_per_shard, 1,
                         "max_sessions_per_shard (1)")
        self.assertEqual(pool.stmtcachesize, 25, "stmtcachesize (25)")
        self.assertEqual(pool.ping_interval, 25, "ping_interval (25)")

    def test_2418_deprecations(self):
        "2418 - test to verify deprecations"
        callback = "pkg_SessionCallback.TheCallback"
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2014:",
                               test_env.get_pool, min=1, max=2, increment=1,
                               wait_timeout=10, waitTimeout=10)
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2014:",
                               test_env.get_pool, min=1, max=2, increment=1,
                               max_lifetime_session=20, maxLifetimeSession=20)
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2014:",
                               test_env.get_pool, min=1, max=2, increment=1,
                               max_sessions_per_shard=1, maxSessionsPerShard=1)
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2014:",
                               test_env.get_pool, min=2, max=8, increment=3,
                               getmode=oracledb.POOL_GETMODE_NOWAIT,
                               session_callback=callback,
                               sessionCallback=callback)

    def test_2419_statement_cache_size(self):
        "2419 - test to verify statement cache size is retained"
        pool = test_env.get_pool(min=1, max=2, increment=1,
                                 getmode=oracledb.POOL_GETMODE_WAIT,
                                 stmtcachesize=25)
        self.assertEqual(pool.stmtcachesize, 25, "stmtcachesize (25)")
        pool.stmtcachesize = 35
        self.assertEqual(pool.stmtcachesize, 35, "stmtcachesize (35)")

    @unittest.skipIf(test_env.get_is_thin(),
                     "thin mode doesn't support tagging yet")
    def test_2420_callable_session_callbacks(self):
        "2420 - test that session callbacks are being called correctly"
        pool = test_env.get_pool(min=2, max=5, increment=1,
                        session_callback=self.__callable_session_callback)

        # new connection with a tag should invoke the session callback
        with pool.acquire(tag="NLS_DATE_FORMAT=SIMPLE") as conn:
            cursor = conn.cursor()
            cursor.execute("select to_char(2021-05-20) from dual")
            result, = cursor.fetchone()
            self.assertEqual(self.session_called, True)

        # acquiring a connection with the same tag should not invoke the
        # session callback
        self.session_called = False
        with pool.acquire(tag="NLS_DATE_FORMAT=SIMPLE") as conn:
            cursor = conn.cursor()
            cursor.execute("select to_char(2021-05-20) from dual")
            result, = cursor.fetchone()
            self.assertEqual(self.session_called, False)

        # acquiring a connection with a new tag should invoke the session
        # callback
        self.session_called = False
        with pool.acquire(tag="NLS_DATE_FORMAT=FULL;TIME_ZONE=UTC") as conn:
            cursor = conn.cursor()
            cursor.execute("select to_char(current_date) from dual")
            result, = cursor.fetchone()
            self.assertEqual(self.session_called, True)

        # acquiring a connection with a new tag and specifying that a
        # connection with any tag can be acquired should invoke the session
        # callback
        self.session_called = False
        with pool.acquire(tag="NLS_DATE_FORMAT=FULL;TIME_ZONE=MST", \
                          matchanytag=True) as conn:
            cursor = conn.cursor()
            cursor.execute("select to_char(current_date) from dual")
            result, = cursor.fetchone()
            self.assertEqual(self.session_called, True)

        # new connection with no tag should invoke the session callback
        self.session_called = False
        with pool.acquire() as conn:
            cursor = conn.cursor()
            cursor.execute("select to_char(current_date) from dual")
            result, = cursor.fetchone()
            self.assertEqual(self.session_called, True)

    def test_2421_pool_close_normal_no_connections(self):
        "2421 - test closing a pool normally with no connections checked out"
        pool = test_env.get_pool(min=1, max=8, increment=1,
                                 getmode=oracledb.POOL_GETMODE_WAIT)
        pool.close()

    def test_2422_pool_close_normal_with_connections(self):
        "2422 - test closing a pool normally with connections checked out"
        pool = test_env.get_pool(min=1, max=8, increment=1,
                                 getmode=oracledb.POOL_GETMODE_WAIT)
        conn = pool.acquire()
        self.assertRaisesRegex(oracledb.InterfaceError, "^DPY-1005:",
                               pool.close)

    def test_2423_pool_close_force(self):
        "2423 - test closing a pool forcibly"
        pool = test_env.get_pool(min=1, max=8, increment=1,
                                 getmode=oracledb.POOL_GETMODE_WAIT)
        conn = pool.acquire()
        pool.close(force=True)

    def test_2424_exception_on_acquire_after_pool_closed(self):
        "2424 - using the pool after it is closed raises an exception"
        pool = test_env.get_pool(min=1, max=8, increment=1,
                                 getmode=oracledb.POOL_GETMODE_WAIT)
        pool.close()
        self.assertRaisesRegex(oracledb.InterfaceError, "^DPY-1002:",
                               pool.acquire)

    def test_2425_pool_with_no_connections(self):
        "2425 - using the pool beyond max limit raises an error"
        pool = test_env.get_pool(min=1, max=2, increment=1,
                                 getmode=oracledb.POOL_GETMODE_WAIT)
        connection_1 = pool.acquire()
        connection_2 = pool.acquire()
        pool.getmode = oracledb.POOL_GETMODE_NOWAIT
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-4005:",
                               pool.acquire)

    def test_2426_session_callback_for_new_connections(self):
        "2426 - callable session callback is executed for new connections"
        class Counter:
            num_calls = 0
            @classmethod
            def session_callback(cls, conn, requested_tag):
                cls.num_calls += 1
        pool = test_env.get_pool(min=1, max=2, increment=1,
                                 session_callback=Counter.session_callback)
        with pool.acquire() as conn1:
            with pool.acquire() as conn2:
                pass
        with pool.acquire() as conn1:
            with pool.acquire() as conn2:
                pass
        self.assertEqual(Counter.num_calls, 2)

    def test_2427_drop_dead_connection_from_pool(self):
        "2427 - drop the pooled connection on receiving dead connection error"
        admin_conn = test_env.get_admin_connection()
        pool = test_env.get_pool(min=2, max=2, increment=2)

        # acquire connections from the pool and kill all the sessions
        with admin_conn.cursor() as admin_cursor:
            for conn in [pool.acquire() for i in range(2)]:
                with conn.cursor() as cursor:
                    cursor.execute("""
                        select
                            dbms_debug_jdwp.current_session_id,
                            dbms_debug_jdwp.current_session_serial
                        from dual""")
                    sid, serial = cursor.fetchone()
                    sql = f"alter system kill session '{sid},{serial}'"
                    admin_cursor.execute(sql)
                conn.close()
        self.assertEqual(pool.opened, 2)

        # when try to re-use the killed sessions error will be raised;
        # release all such connections
        for conn in [pool.acquire() for i in range(2)]:
            with conn.cursor() as cursor:
                self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-4011:",
                                       cursor.execute, "select user from dual")
            conn.close()
        self.assertEqual(pool.opened, 0)

        # if a free connection is available, it can be used; otherwise a new
        # connection will be created
        for conn in [pool.acquire() for i in range(2)]:
            with conn.cursor() as cursor:
                cursor.execute("select user from dual")
                user, = cursor.fetchone()
                self.assertEqual(user, test_env.get_main_user().upper())
            conn.close()
        self.assertEqual(pool.opened, 2)

    def test_2428_acquire_conn_from_empty_pool(self):
        "2428 - acquire a connection from an empty pool (min=0)"
        pool = test_env.get_pool(min=0, max=2, increment=2)
        with pool.acquire() as conn:
            with conn.cursor() as cursor:
                cursor.execute("select user from dual")
                result, = cursor.fetchone()
                self.assertEqual(result, test_env.get_main_user().upper())

    @unittest.skipIf(test_env.get_is_thin(),
                     "thin mode doesn't support soda_metadata_cache" \
                     "parameter yet")
    def test_2429_soda_metadata_cache(self):
        "2429 - test soda_metadata_cache parameter"
        self.get_soda_database(minclient=(19, 11))
        pool = test_env.get_pool(min=0, max=2, increment=2)
        self.assertEqual(pool.soda_metadata_cache, False)
        pool = test_env.get_pool(min=0, max=2, increment=2,
                                 soda_metadata_cache=True)
        self.assertEqual(pool.soda_metadata_cache, True)
        pool.soda_metadata_cache = False
        self.assertEqual(pool.soda_metadata_cache, False)

    def test_2430_get_different_types_from_pooled_connections(self):
        "2430 - get different object types from different connections"
        pool = test_env.get_pool(min=1, max=2, increment=1)
        with pool.acquire() as conn:
            typ = conn.gettype("UDT_SUBOBJECT")
            self.assertEqual(typ.name, "UDT_SUBOBJECT")
        with pool.acquire() as conn:
            typ = conn.gettype("UDT_OBJECTARRAY")
            self.assertEqual(typ.name, "UDT_OBJECTARRAY")

    def test_2431_proxy_user_in_create(self):
        "2431 - test creating a pool using a proxy user"
        user_str = "%s[%s]" % \
                (test_env.get_main_user(), test_env.get_proxy_user())
        pool = test_env.get_pool(user=user_str)
        self.__verify_connection(pool.acquire(), test_env.get_proxy_user(),
                                 test_env.get_main_user())

    def test_2432_conn_acquire_in_lifo(self):
        "2432 - test acquiring conn from pool in LIFO order"
        pool = test_env.get_pool(min=5, max=10, increment=1,
                                 getmode=oracledb.POOL_GETMODE_WAIT)
        sql = "select sys_context('userenv', 'sid') from dual"
        conns = [pool.acquire() for i in range(3)]
        sids = [c.cursor().execute(sql).fetchone()[0] for c in conns]

        conns[1].close()
        conns[2].close()
        conns[0].close()

        conn = pool.acquire()
        sid = conn.cursor().execute(sql).fetchone()[0]
        self.assertEqual(sid, sids[0], "not LIFO")

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2021, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# subscr.py
#
# Contains the Subscription class and Message classes used for managing
# subscriptions to database events and the messages that are sent when those
# events are detected.
#------------------------------------------------------------------------------

from typing import Callable, Union, List
from . import connection

class Subscription:

    def __repr__(self):
        return f"<oracledb.Subscription on {self.connection!r}>"

    @classmethod
    def _from_impl(cls, impl):
        subscr = cls.__new__(cls)
        subscr._impl = impl
        return subscr

    @property
    def callback(self) -> Callable:
        """
        Returns the callback that was registered when the subscription was
        created.
        """
        return self._impl.callback

    @property
    def connection(self) -> "connection.Connection":
        """
        Returns the connection that was used to register the subscription when
        it was created.
        """
        return self._impl.connection

    @property
    def id(self) -> int:
        """
        Returns the value of REGID found in the database view
        USER_CHANGE_NOTIFICATION_REGS or the value of REG_ID found in the
        database view USER_SUBSCR_REGISTRATIONS. For AQ subscriptions, this
        value is 0.
        """
        return self._impl.id

    @property
    def ip_address(self) -> str:
        """
        Returns the IP address used for callback notifications from the
        database server. If not set during construction, this value is None.
        """
        return self._impl.ip_address

    @property
    def ipAddress(self) -> str:
        """
        Deprecated. Use property ip_address instead.
        """
        return self.ip_address

    @property
    def name(self) -> str:
        """
        Returns the name used to register the subscription when it was created.
        """
        return self._impl.name

    @property
    def namespace(self) -> int:
        """
        Returns the namespace used to register the subscription when it was
        created.
        """
        return self._impl.namespace

    @property
    def operations(self) -> int:
        """
        Returns the operations that will send notifications for each table or
        query that is registered using this subscription.
        """
        return self._impl.operations

    @property
    def port(self) -> int:
        """
        Returns the port used for callback notifications from the database
        server. If not set during construction, this value is zero.
        """
        return self._impl.port

    @property
    def protocol(self) -> int:
        """
        Returns the protocol used to register the subscription when it was
        created.
        """
        return self._impl.protocol

    @property
    def qos(self) -> int:
        """
        Returns the quality of service flags used to register the subscription
        when it was created.
        """
        return self._impl.qos

    def registerquery(self, statement: str,
                      args: Union[list, dict]=None) -> int:
        """
        Register the query for subsequent notification when tables referenced
        by the query are changed. This behaves similarly to cursor.execute()
        but only queries are permitted and the args parameter, if specified,
        must be a sequence or dictionary. If the qos parameter included the
        flag SUBSCR_QOS_QUERY when the subscription was created, then the ID
        for the registered query is returned; otherwise, None is returned.
        """
        if args is not None and not isinstance(args, (list, dict)):
            raise TypeError("expecting args to be a dictionary or list")
        return self._impl.register_query(statement, args)

    @property
    def timeout(self) -> int:
        """
        Returns the timeout (in seconds) that was specified when the
        subscription was created. A value of 0 indicates that there is no
        timeout.
        """
        return self._impl.timeout


class Message:

    def __init__(self, subscription: Subscription) -> None:
        self._subscription = subscription
        self._consumer_name = None
        self._dbname = None
        self._queries = []
        self._queue_name = None
        self._registered = False
        self._tables = []
        self._txid = None
        self._type = 0
        self._msgid = None


    @property
    def consumer_name(self) -> Union[str, None]:
        """
        Returns the name of the consumer which generated the notification. It
        will be populated if the subscription was created with the namespace
        SUBSCR_NAMESPACE_AQ and the queue is a multiple consumer queue.
        """
        return self._consumer_name

    @property
    def consumerName(self) -> Union[str, None]:
        """
        Deprecated. Use property consumer_name instead.
        """
        return self.consumer_name

    @property
    def dbname(self) -> Union[str, None]:
        """
        Returns the name of the database that generated the notification.
        """
        return self._dbname

    @property
    def msgid(self) -> Union[bytes, None]:
        """
        Returns the message id of the AQ message that generated the
        notification.
        """
        return self._msgid

    @property
    def queries(self) -> List["MessageQuery"]:
        """
        Returns a list of message query objects that give information about
        query result sets changed for this notification. This attribute will be
        an empty list if the qos parameter did not include the flag
        SUBSCR_QOS_QUERY when the subscription was created.
        """
        return self._queries

    @property
    def queue_name(self) -> Union[str, None]:
        """
        Returns the name of the queue which generated the notification. It will
        only be populated if the subscription was created with the namespace
        SUBSCR_NAMESPACE_AQ.
        """
        return self._queue_name

    @property
    def queueName(self) -> Union[str, None]:
        """
        Deprecated. Use property queue_name instead.
        """
        return self.queue_name

    @property
    def registered(self) -> bool:
        """
        Returns whether the subscription which generated this notification is
        still registered with the database. The subscription is automatically
        deregistered with the database when the subscription timeout value is
        reached or when the first notification is sent (when the quality of
        service flag SUBSCR_QOS_DEREG_NFY is used).
        """
        return self._registered

    @property
    def subscription(self) -> Subscription:
        """
        Returns the subscription object for which this notification was
        generated.
        """
        return self._subscription

    @property
    def tables(self) -> List["MessageTable"]:
        """
        Returns a list of message table objects that give information about the
        tables changed for this notification. This attribute will be an empty
        list if the qos parameter included the flag SUBSCR_QOS_QUERY when the
        subscription was created.
        """
        return self._tables

    @property
    def txid(self) -> Union[bytes, None]:
        """
        Returns the id of the transaction that generated the notification.
        """
        return self._txid

    @property
    def type(self) -> int:
        """
        Returns the type of message that has been sent.
        """
        return self._type


class MessageQuery:

    def __init__(self) -> None:
        self._id = 0
        self._operation = 0
        self._tables = []

    @property
    def id(self) -> int:
        """
        Returns the query id of the query for which the result set changed. The
        value will match the value returned by Subscription.registerquery()
        when the related query was registered.
        """
        return self._id

    @property
    def operation(self) -> int:
        """
        Returns the operation that took place on the query result set that was
        changed. Valid values for this attribute are EVENT_DEREG and
        EVENT_QUERYCHANGE.
        """
        return self._operation

    @property
    def tables(self) -> List["MessageTable"]:
        """
        Returns a list of message table objects that give information about the
        table changes that caused the query result set to change for this
        notification.
        """
        return self._tables


class MessageRow:

    def __init__(self) -> None:
        self._operation = 0
        self._rowid = None

    @property
    def operation(self) -> int:
        """
        Returns the operation that took place on the row that was changed.
        """
        return self._operation

    @property
    def rowid(self) -> Union[str, None]:
        """
        Returns the rowid of the row that was changed.
        """
        return self._rowid


class MessageTable:

    def __init__(self) -> None:
        self._name = None
        self._operation = 0
        self._rows = []

    @property
    def name(self) -> Union[str, None]:
        """
        Returns the name of the table that was changed.
        """
        return self._name

    @property
    def operation(self) -> int:
        """
        Returns the operation that took place on the table that was changed.
        """
        return self._operation

    @property
    def rows(self) -> List["MessageRow"]:
        """
        Returns a list of message row objects that give information about the
        rows changed on the table. This value is only filled in if the qos
        parameter to the Connection.subscribe() method included the flag
        SUBSCR_QOS_ROWIDS.
        """
        return self._rows



#------------------------------------------------------------------------------
# Copyright (c) 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# bind_var.pyx
#
# Cython file defining the BindVar implementation class (embedded in
# base_impl.pyx).
#------------------------------------------------------------------------------

@cython.freelist(20)
cdef class BindVar:

    cdef int _create_var_from_type(self, object conn,
                                   BaseCursorImpl cursor_impl,
                                   object value) except -1:
        """
        Creates a variable given type information. This may be supplied as a
        list of size 2 (type, num_elements) which internally creates an array
        variable, or as an integer which internally creates a string of the
        given length, or as a database type, API type or Python type.
        """
        cdef BaseVarImpl var_impl
        var_impl = cursor_impl._create_var_impl(conn)
        var_impl.num_elements = 1
        if isinstance(value, list):
            if len(value) != 2:
                errors._raise_err(errors.ERR_WRONG_ARRAY_DEFINITION)
            var_impl._set_type_info_from_type(value[0])
            var_impl.num_elements = value[1]
            var_impl.is_array = True
        elif isinstance(value, int):
            var_impl.dbtype = DB_TYPE_VARCHAR
            var_impl.size = value
        else:
            var_impl._set_type_info_from_type(value)
        var_impl._finalize_init()
        self.var_impl = var_impl
        if isinstance(value, PY_TYPE_DB_OBJECT_TYPE):
            self.var = PY_TYPE_VAR._from_impl(self.var_impl, value)

    cdef int _create_var_from_value(self, object conn,
                                    BaseCursorImpl cursor_impl, object value,
                                    uint32_t num_elements) except -1:
        """
        Creates a variable using the value as a template.
        """
        cdef:
            bint is_plsql = cursor_impl._is_plsql()
            BaseVarImpl var_impl
        var_impl = cursor_impl._create_var_impl(conn)
        if not isinstance(value, list):
            var_impl.num_elements = num_elements
            var_impl._set_type_info_from_value(value, is_plsql)
        else:
            var_impl.is_array = True
            var_impl.num_elements = max(num_elements, len(value))
            for element in value:
                if element is not None:
                    var_impl._set_type_info_from_value(element, is_plsql)
            if var_impl.dbtype is None:
                var_impl.dbtype = DB_TYPE_VARCHAR
                var_impl.size = 1
        var_impl._finalize_init()
        self.var_impl = var_impl

    cdef int _set_by_type(self, object conn, BaseCursorImpl cursor_impl,
                          object typ) except -1:
        """
        Sets the bind variable information given a type.
        """
        if typ is not None:
            if isinstance(typ, PY_TYPE_VAR):
                self.var = typ
                self.var_impl = typ._impl
            else:
                self._create_var_from_type(conn, cursor_impl, typ)

    cdef int _set_by_value(self, object conn, BaseCursorImpl cursor_impl,
                           object cursor, object value, object type_handler,
                           uint32_t row_num, uint32_t num_elements,
                           bint defer_type_assignment) except -1:
        """
        Sets the bind variable information given a value. The row number
        supplied is used as the offset into the variable value array. Type
        assignment is deferred for None values if specified. Once a value that
        is not None is set, an exception is raised for any non-compliant values
        that are seen after that.
        """
        cdef:
            bint was_set = True
            object var

        # a variable can be set directly in which case nothing further needs to
        # be done!
        if isinstance(value, PY_TYPE_VAR):
            if value is not self.var:
                self.var = value
                self.var_impl = value._impl
            return 0

        # if a variable already exists check to see if the value can be set on
        # that variable; an exception is raised if a value has been previously
        # set on that bind variable; otherwise, the variable is replaced with a
        # new one
        if self.var_impl is not None:
            if self.has_value:
                return self.var_impl._check_and_set_value(row_num, value, NULL)
            self.var_impl._check_and_set_value(row_num, value, &was_set)
            if was_set:
                self.has_value = True
                return 0
            self.var_impl = None
            self.var = None

        # a new variable needs to be created; if the value is null (None),
        # however, and type assignment is deferred, nothing to do!
        if value is None and defer_type_assignment:
            return 0

        # if an input type handler is specified, call it; the input type
        # handler should return a variable or None; the value None implies
        # that the default processing should take place just as if no input
        # type handler was defined
        if type_handler is not None:
            var = type_handler(cursor, value, num_elements)
            if var is not None:
                if not isinstance(var, PY_TYPE_VAR):
                    errors._raise_err(errors.ERR_EXPECTING_VAR)
                self.var = var
                self.var_impl = var._impl
                self.var_impl._check_and_set_value(row_num, value, NULL)
                self.has_value = True
                return 0

        # otherwise, if no type handler exists or the type handler returned
        # the value None, create a new variable deriving type information
        # from the value that is being set
        self._create_var_from_value(conn, cursor_impl, value, num_elements)
        self.var_impl._check_and_set_value(row_num, value, NULL)
        self.has_value = True

    def get_value(self, uint32_t pos):
        """
        Internal method for getting the value of a variable at a given
        position.
        """
        if self.var_impl is not None:
            return self.var_impl.get_value(pos)



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2023, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
3200 - Module for testing features introduced in 12.1
"""

import datetime
import unittest

import oracledb
import test_env

@unittest.skipUnless(test_env.get_client_version() >= (12, 1),
                     "unsupported client")
class TestCase(test_env.BaseTestCase):

    def test_3200_array_dml_row_counts_off(self):
        "3200 - test executing with arraydmlrowcounts mode disabled"
        self.cursor.execute("truncate table TestArrayDML")
        rows = [(1, "First"), (2, "Second")]
        sql = "insert into TestArrayDML (IntCol,StringCol) values (:1,:2)"
        self.cursor.executemany(sql, rows, arraydmlrowcounts=False)
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-4006:",
                               self.cursor.getarraydmlrowcounts)
        rows = [(3, "Third"), (4, "Fourth")]
        self.cursor.executemany(sql, rows)
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-4006:",
                               self.cursor.getarraydmlrowcounts)

    def test_3201_array_dml_row_counts_on(self):
        "3201 - test executing with arraydmlrowcounts mode enabled"
        self.cursor.execute("truncate table TestArrayDML")
        rows = [
            (1, "First", 100),
            (2, "Second", 200),
            (3, "Third", 300),
            (4, "Fourth", 300),
            (5, "Fifth", 300)
        ]
        sql = "insert into TestArrayDML (IntCol,StringCol,IntCol2) " \
              "values (:1,:2,:3)"
        self.cursor.executemany(sql, rows, arraydmlrowcounts=True)
        self.connection.commit()
        self.assertEqual(self.cursor.getarraydmlrowcounts(), [1, 1, 1, 1, 1])
        self.cursor.execute("select count(*) from TestArrayDML")
        count, = self.cursor.fetchone()
        self.assertEqual(count, len(rows))

    def test_3202_bind_plsql_boolean_collection_in(self):
        "3202 - test binding a boolean collection (in)"
        type_obj = self.connection.gettype("PKG_TESTBOOLEANS.UDT_BOOLEANLIST")
        obj = type_obj.newobject()
        obj.setelement(1, True)
        obj.extend([True, False, True, True, False, True])
        result = self.cursor.callfunc("pkg_TestBooleans.TestInArrays", int,
                                      (obj,))
        self.assertEqual(result, 5)

    def test_3203_bind_plsql_boolean_collection_out(self):
        "3203 - test binding a boolean collection (out)"
        type_obj = self.connection.gettype("PKG_TESTBOOLEANS.UDT_BOOLEANLIST")
        obj = type_obj.newobject()
        self.cursor.callproc("pkg_TestBooleans.TestOutArrays", (6, obj))
        self.assertEqual(obj.aslist(), [True, False, True, False, True, False])

    def test_3204_bind_plql_date_collection_in(self):
        "3204 - test binding a PL/SQL date collection (in)"
        type_obj = self.connection.gettype("PKG_TESTDATEARRAYS.UDT_DATELIST")
        obj = type_obj.newobject()
        obj.setelement(1, datetime.datetime(2016, 2, 5))
        obj.append(datetime.datetime(2016, 2, 8, 12, 15, 30))
        obj.append(datetime.datetime(2016, 2, 12, 5, 44, 30))
        result = self.cursor.callfunc("pkg_TestDateArrays.TestInArrays",
                                      oracledb.NUMBER,
                                      (2, datetime.datetime(2016, 2, 1), obj))
        self.assertEqual(result, 24.75)

    def test_3205_bind_plqsl_date_collection_in_out(self):
        "3205 - test binding a PL/SQL date collection (in/out)"
        type_obj = self.connection.gettype("PKG_TESTDATEARRAYS.UDT_DATELIST")
        obj = type_obj.newobject()
        obj.setelement(1, datetime.datetime(2016, 1, 1))
        obj.append(datetime.datetime(2016, 1, 7))
        obj.append(datetime.datetime(2016, 1, 13))
        obj.append(datetime.datetime(2016, 1, 19))
        self.cursor.callproc("pkg_TestDateArrays.TestInOutArrays", (4, obj))
        expected_values = [
            datetime.datetime(2016, 1, 8),
            datetime.datetime(2016, 1, 14),
            datetime.datetime(2016, 1, 20),
            datetime.datetime(2016, 1, 26)
        ]
        self.assertEqual(obj.aslist(), expected_values)

    def test_3206_bind_plsql_date_collection_out(self):
        "3206 - test binding a PL/SQL date collection (out)"
        type_obj = self.connection.gettype("PKG_TESTDATEARRAYS.UDT_DATELIST")
        obj = type_obj.newobject()
        self.cursor.callproc("pkg_TestDateArrays.TestOutArrays", (3, obj))
        expected_values = [
            datetime.datetime(2002, 12, 13, 4, 48),
            datetime.datetime(2002, 12, 14, 9, 36),
            datetime.datetime(2002, 12, 15, 14, 24)
        ]
        self.assertEqual(obj.aslist(), expected_values)

    def test_3207_bind_plsql_number_collection_in(self):
        "3207 - test binding a PL/SQL number collection (in)"
        type_name = "PKG_TESTNUMBERARRAYS.UDT_NUMBERLIST"
        type_obj = self.connection.gettype(type_name)
        obj = type_obj.newobject()
        obj.setelement(1, 10)
        obj.extend([20, 30, 40, 50])
        result = self.cursor.callfunc("pkg_TestNumberArrays.TestInArrays", int,
                                      (5, obj))
        self.assertEqual(result, 155)

    def test_3208_bind_plsql_number_collection_in_out(self):
        "3208 - test binding a PL/SQL number collection (in/out)"
        type_name = "PKG_TESTNUMBERARRAYS.UDT_NUMBERLIST"
        type_obj = self.connection.gettype(type_name)
        obj = type_obj.newobject()
        obj.setelement(1, 5)
        obj.extend([8, 3, 2])
        self.cursor.callproc("pkg_TestNumberArrays.TestInOutArrays", (4, obj))
        self.assertEqual(obj.aslist(), [50, 80, 30, 20])

    def test_3209_bind_plsql_number_collection_out(self):
        "3209 - test binding a PL/SQL number collection (out)"
        type_name = "PKG_TESTNUMBERARRAYS.UDT_NUMBERLIST"
        type_obj = self.connection.gettype(type_name)
        obj = type_obj.newobject()
        self.cursor.callproc("pkg_TestNumberArrays.TestOutArrays", (3, obj))
        self.assertEqual(obj.aslist(), [100, 200, 300])

    def test_3210_bind_plsql_record_array(self):
        "3210 - test binding an array of PL/SQL records (in)"
        rec_type = self.connection.gettype("PKG_TESTRECORDS.UDT_RECORD")
        array_type = self.connection.gettype("PKG_TESTRECORDS.UDT_RECORDARRAY")
        array_obj = array_type.newobject()
        for i in range(3):
            obj = rec_type.newobject()
            obj.NUMBERVALUE = i + 1
            obj.STRINGVALUE = "String in record #%d" % (i + 1)
            obj.DATEVALUE = datetime.datetime(2017, i + 1, 1)
            obj.TIMESTAMPVALUE = datetime.datetime(2017, 1, i + 1)
            obj.BOOLEANVALUE = (i % 2) == 1
            obj.PLSINTEGERVALUE = i * 5
            obj.BINARYINTEGERVALUE = i * 2
            array_obj.append(obj)
        result = self.cursor.callfunc("pkg_TestRecords.TestInArrays", str,
                                      (array_obj,))
        self.assertEqual(result,
                         "udt_Record(1, 'String in record #1', " \
                         "to_date('2017-01-01', 'YYYY-MM-DD'), " \
                         "to_timestamp('2017-01-01 00:00:00', " \
                         "'YYYY-MM-DD HH24:MI:SS'), false, 0, 0); " \
                         "udt_Record(2, 'String in record #2', " \
                         "to_date('2017-02-01', 'YYYY-MM-DD'), " \
                         "to_timestamp('2017-01-02 00:00:00', " \
                         "'YYYY-MM-DD HH24:MI:SS'), true, 5, 2); " \
                         "udt_Record(3, 'String in record #3', " \
                         "to_date('2017-03-01', 'YYYY-MM-DD'), " \
                         "to_timestamp('2017-01-03 00:00:00', " \
                         "'YYYY-MM-DD HH24:MI:SS'), false, 10, 4)")

    def test_3211_bind_plsql_record_in(self):
        "3211 - test binding a PL/SQL record (in)"
        type_obj = self.connection.gettype("PKG_TESTRECORDS.UDT_RECORD")
        obj = type_obj.newobject()
        obj.NUMBERVALUE = 18
        obj.STRINGVALUE = "A string in a record"
        obj.DATEVALUE = datetime.datetime(2016, 2, 15)
        obj.TIMESTAMPVALUE = datetime.datetime(2016, 2, 12, 14, 25, 36)
        obj.BOOLEANVALUE = False
        obj.PLSINTEGERVALUE = 21
        obj.BINARYINTEGERVALUE = 5
        result = self.cursor.callfunc("pkg_TestRecords.GetStringRep", str,
                                      (obj,))
        self.assertEqual(result,
                         "udt_Record(18, 'A string in a record', " \
                         "to_date('2016-02-15', 'YYYY-MM-DD'), " \
                         "to_timestamp('2016-02-12 14:25:36', " \
                         "'YYYY-MM-DD HH24:MI:SS'), false, 21, 5)")

    def test_3212_bind_plsql_record_out(self):
        "3212 - test binding a PL/SQL record (out)"
        type_obj = self.connection.gettype("PKG_TESTRECORDS.UDT_RECORD")
        obj = type_obj.newobject()
        obj.NUMBERVALUE = 5
        obj.STRINGVALUE = "Test value"
        obj.DATEVALUE = datetime.datetime.today()
        obj.TIMESTAMPVALUE = datetime.datetime.today()
        obj.BOOLEANVALUE = False
        obj.PLSINTEGERVALUE = 23
        obj.BINARYINTEGERVALUE = 9
        self.cursor.callproc("pkg_TestRecords.TestOut", (obj,))
        self.assertEqual(obj.NUMBERVALUE, 25)
        self.assertEqual(obj.STRINGVALUE, "String in record")
        self.assertEqual(obj.DATEVALUE, datetime.datetime(2016, 2, 16))
        self.assertEqual(obj.TIMESTAMPVALUE,
                         datetime.datetime(2016, 2, 16, 18, 23, 55))
        self.assertEqual(obj.BOOLEANVALUE, True)
        self.assertEqual(obj.PLSINTEGERVALUE, 45)
        self.assertEqual(obj.BINARYINTEGERVALUE, 10)

    def test_3213_bind_plsql_string_collection_in(self):
        "3213 - test binding a PL/SQL string collection (in)"
        type_name = "PKG_TESTSTRINGARRAYS.UDT_STRINGLIST"
        type_obj = self.connection.gettype(type_name)
        obj = type_obj.newobject()
        obj.setelement(1, "First element")
        obj.setelement(2, "Second element")
        obj.setelement(3, "Third element")
        result = self.cursor.callfunc("pkg_TestStringArrays.TestInArrays", int,
                                      (5, obj))
        self.assertEqual(result, 45)

    def test_3214_bind_plsql_string_collection_in_out(self):
        "3214 - test binding a PL/SQL string collection (in/out)"
        type_name = "PKG_TESTSTRINGARRAYS.UDT_STRINGLIST"
        type_obj = self.connection.gettype(type_name)
        obj = type_obj.newobject()
        obj.setelement(1, "The first element")
        obj.append("The second element")
        obj.append("The third and final element")
        self.cursor.callproc("pkg_TestStringArrays.TestInOutArrays", (3, obj))
        expected_values = [
            'Converted element # 1 originally had length 17',
            'Converted element # 2 originally had length 18',
            'Converted element # 3 originally had length 27'
        ]
        self.assertEqual(obj.aslist(), expected_values)

    def test_3215_bind_plsql_string_collection_out(self):
        "3215 - test binding a PL/SQL string collection (out)"
        type_name = "PKG_TESTSTRINGARRAYS.UDT_STRINGLIST"
        type_obj = self.connection.gettype(type_name)
        obj = type_obj.newobject()
        self.cursor.callproc("pkg_TestStringArrays.TestOutArrays", (4, obj))
        expected_values = [
            'Test out element # 1',
            'Test out element # 2',
            'Test out element # 3',
            'Test out element # 4'
        ]
        self.assertEqual(obj.aslist(), expected_values)

    def test_3216_bind_plsql_string_collection_out_with_holes(self):
        "3216 - test binding a PL/SQL string collection (out with holes)"
        type_name = "PKG_TESTSTRINGARRAYS.UDT_STRINGLIST"
        type_obj = self.connection.gettype(type_name)
        obj = type_obj.newobject()
        self.cursor.callproc("pkg_TestStringArrays.TestIndexBy", (obj,))
        self.assertEqual(obj.first(), -1048576)
        self.assertEqual(obj.last(), 8388608)
        self.assertEqual(obj.next(-576), 284)
        self.assertEqual(obj.prev(284), -576)
        self.assertEqual(obj.size(), 4)
        self.assertEqual(obj.exists(-576), True)
        self.assertEqual(obj.exists(-577), False)
        self.assertEqual(obj.getelement(284), 'Third element')
        expected_list = [
            "First element",
            "Second element",
            "Third element",
            "Fourth element"
        ]
        self.assertEqual(obj.aslist(), expected_list)
        expected_dict = {
            -1048576: 'First element',
            -576: 'Second element',
            284: 'Third element',
            8388608: 'Fourth element'
        }
        self.assertEqual(obj.asdict(), expected_dict)
        obj.delete(-576)
        obj.delete(284)
        expected_list.pop(2)
        expected_list.pop(1)
        self.assertEqual(obj.aslist(), expected_list)
        expected_dict.pop(-576)
        expected_dict.pop(284)
        self.assertEqual(obj.asdict(), expected_dict)

    def test_3217_exception_in_iteration(self):
        "3217 - test executing with arraydmlrowcounts with exception"
        self.cursor.execute("truncate table TestArrayDML")
        rows = [
            (1, "First"),
            (2, "Second"),
            (2, "Third"),
            (4, "Fourth")
        ]
        sql = "insert into TestArrayDML (IntCol,StringCol) values (:1,:2)"
        self.assertRaisesRegex(oracledb.DatabaseError, "^ORA-00001:",
                               self.cursor.executemany, sql, rows,
                               arraydmlrowcounts=True)
        self.assertEqual(self.cursor.getarraydmlrowcounts(), [1, 1])

    def test_3218_executing_delete(self):
        "3218 - test executing delete statement with arraydmlrowcount mode"
        self.cursor.execute("truncate table TestArrayDML")
        rows = [
            (1, "First", 100),
            (2, "Second", 200),
            (3, "Third", 300),
            (4, "Fourth", 300),
            (5, "Fifth", 300),
            (6, "Sixth", 400),
            (7, "Seventh", 400),
            (8, "Eighth", 500)
        ]
        sql = "insert into TestArrayDML (IntCol,StringCol,IntCol2) " \
              "values (:1, :2, :3)"
        self.cursor.executemany(sql, rows)
        rows = [(200,), (300,), (400,)]
        statement = "delete from TestArrayDML where IntCol2 = :1"
        self.cursor.executemany(statement, rows, arraydmlrowcounts=True)
        self.assertEqual(self.cursor.getarraydmlrowcounts(), [1, 3, 2])
        self.assertEqual(self.cursor.rowcount, 6)

    def test_3219_executing_update(self):
        "3219 - test executing update statement with arraydmlrowcount mode"
        self.cursor.execute("truncate table TestArrayDML")
        rows = [
            (1, "First",100),
            (2, "Second",200),
            (3, "Third",300),
            (4, "Fourth",300),
            (5, "Fifth",300),
            (6, "Sixth",400),
            (7, "Seventh",400),
            (8, "Eighth",500)
        ]
        sql = "insert into TestArrayDML (IntCol,StringCol,IntCol2) " \
              "values (:1, :2, :3)"
        self.cursor.executemany(sql, rows)
        rows = [
            ("One", 100),
            ("Two", 200),
            ("Three", 300),
            ("Four", 400)
        ]
        sql = "update TestArrayDML set StringCol = :1 where IntCol2 = :2"
        self.cursor.executemany(sql, rows, arraydmlrowcounts=True)
        self.assertEqual(self.cursor.getarraydmlrowcounts(), [1, 1, 3, 2])
        self.assertEqual(self.cursor.rowcount, 7)

    def test_3220_implicit_results(self):
        "3220 - test getimplicitresults() returns the correct data"
        self.cursor.execute("""
                declare
                    c1 sys_refcursor;
                    c2 sys_refcursor;
                begin
                    open c1 for
                        select NullableCol
                        from TestNumbers
                        where IntCol between 3 and 5;

                    dbms_sql.return_result(c1);

                    open c2 for
                        select NullableCol
                        from TestNumbers
                        where IntCol between 7 and 10;

                    dbms_sql.return_result(c2);
                end;""")
        results = self.cursor.getimplicitresults()
        self.assertEqual(len(results), 2)
        self.assertEqual([n for n, in results[0]],
                         [2924207, None, 59797108943])
        self.assertEqual([n for n, in results[1]],
                         [1222791080775407, None, 25004854810776297743, None])

    def test_3221_implicit_results_no_statement(self):
        "3221 - test getimplicitresults() without executing a statement"
        cursor = self.connection.cursor()
        self.assertRaisesRegex(oracledb.InterfaceError, "^DPY-1004:",
                               cursor.getimplicitresults)

    def test_3222_insert_with_batch_error(self):
        "3222 - test executing insert with multiple distinct batch errors"
        self.cursor.execute("truncate table TestArrayDML")
        rows = [
            (1, "First", 100),
            (2, "Second", 200),
            (2, "Third", 300),
            (4, "Fourth", 400),
            (5, "Fourth", 1000)
        ]
        sql = "insert into TestArrayDML (IntCol, StringCol, IntCol2) " \
              "values (:1, :2, :3)"
        self.cursor.executemany(sql, rows, batcherrors=True,
                                arraydmlrowcounts=True)
        user = test_env.get_main_user()
        expected_errors = [
            (4, "ORA-01438"),
            (2, "ORA-00001")
        ]
        actual_errors = [(e.offset, e.full_code) \
                         for e in self.cursor.getbatcherrors()]
        self.assertEqual(actual_errors, expected_errors)
        self.assertEqual(self.cursor.getarraydmlrowcounts(), [1, 1, 0, 1, 0])

    def test_3223_batch_error_false(self):
        "3223 - test batcherrors mode set to False"
        self.cursor.execute("truncate table TestArrayDML")
        rows = [
            (1, "First", 100),
            (2, "Second", 200),
            (2, "Third", 300)
        ]
        sql = "insert into TestArrayDML (IntCol, StringCol, IntCol2) " \
              "values (:1, :2, :3)"
        self.assertRaisesRegex(oracledb.IntegrityError, "^ORA-00001:",
                               self.cursor.executemany, sql, rows,
                               batcherrors=False)

    def test_3224_update_with_batch_error(self):
        "3224 - test executing in succession with batch error"
        self.cursor.execute("truncate table TestArrayDML")
        rows = [
            (1, "First", 100),
            (2, "Second", 200),
            (3, "Third", 300),
            (4, "Second", 300),
            (5, "Fifth", 300),
            (6, "Sixth", 400),
            (6, "Seventh", 400),
            (8, "Eighth", 100)
        ]
        sql = "insert into TestArrayDML (IntCol, StringCol, IntCol2) " \
              "values (:1, :2, :3)"
        self.cursor.executemany(sql, rows, batcherrors=True)
        expected_errors = [(6, "ORA-00001")]
        actual_errors = [(e.offset, e.full_code) \
                         for e in self.cursor.getbatcherrors()]
        self.assertEqual(actual_errors, expected_errors)
        rows = [
            (101, "First"),
            (201, "Second"),
            (3000, "Third"),
            (900, "Ninth"),
            (301, "Third")
        ]
        sql = "update TestArrayDML set IntCol2 = :1 where StringCol = :2"
        self.cursor.executemany(sql, rows, arraydmlrowcounts=True,
                                batcherrors=True)
        expected_errors = [(2, "ORA-01438")]
        actual_errors = [(e.offset, e.full_code) \
                         for e in self.cursor.getbatcherrors()]
        self.assertEqual(actual_errors, expected_errors)
        self.assertEqual(self.cursor.getarraydmlrowcounts(), [1, 2, 0, 0, 1])
        self.assertEqual(self.cursor.rowcount, 4)

    def test_3225_implicit_results(self):
        "3225 - test using implicit cursors to execute new statements"
        cursor = self.connection.cursor()
        cursor.execute("""
                declare
                    c1 sys_refcursor;
                begin
                    open c1 for
                        select NumberCol
                        from TestNumbers
                        where IntCol between 3 and 5;

                    dbms_sql.return_result(c1);
                end;""")
        results = cursor.getimplicitresults()
        self.assertEqual(len(results), 1)
        self.assertEqual([n for n, in results[0]], [3.75, 5, 6.25])
        results[0].execute("select :1 from dual", (7,))
        row, = results[0].fetchone()
        self.assertEqual(row, 7)

    def test_3226_batch_error_no_errors(self):
        "3226 - test batcherrors mode without any errors produced"
        self.cursor.execute("truncate table TestArrayDML")
        rows = [
            (1, "First", 100),
            (2, "Second", 200),
            (3, "Third", 300)
        ]
        sql = """
                insert into TestArrayDML (IntCol, StringCol, IntCol2)
                values (:1, :2, :3)"""
        self.cursor.executemany(sql, rows, batcherrors=True)
        self.assertEqual(self.cursor.getbatcherrors(), [])

    def test_3227_batch_error_multiple_execute(self):
        "3227 - test batcherrors mode with multiple executes"
        self.cursor.execute("truncate table TestArrayDML")
        rows_1 = [
            (1, "Value 1", 100),
            (2, "Value 2", 200),
            (2, "Value 2", 200)
        ]
        rows_2 = [
            (3, "Value 3", 300),
            (3, "Value 3", 300),
            (4, "Value 4", 400)
        ]
        sql = """
                insert into TestArrayDML (IntCol, StringCol, IntCol2)
                values (:1, :2, :3)"""
        self.cursor.executemany(sql, rows_1, batcherrors=True)
        expected_errors = [(2, "ORA-00001")]
        actual_errors = [(e.offset, e.full_code) \
                         for e in self.cursor.getbatcherrors()]
        self.assertEqual(actual_errors, expected_errors)
        self.cursor.executemany(sql, rows_2, batcherrors=True)
        expected_errors = [(1, "ORA-00001")]
        actual_errors = [(e.offset, e.full_code) \
                         for e in self.cursor.getbatcherrors()]
        self.assertEqual(actual_errors, expected_errors)

    def test_3228_record_based_on_table_type(self):
        "3228 - test %ROWTYPE record type"
        type_obj = self.connection.gettype("TESTTEMPTABLE%ROWTYPE")
        self.assertEqual(type_obj.attributes[3].name, "NUMBERCOL")

    def test_3229_collection_of_records_based_on_table_type(self):
        "3229 - test collection of %ROWTYPE record type"
        type_name = "PKG_TESTBINDOBJECT.UDT_COLLECTIONROWTYPE"
        type_obj = self.connection.gettype(type_name)
        self.assertEqual(type_obj.element_type.attributes[3].name, "NUMBERCOL")

    def test_3230_batcherrors_with_plsql(self):
        "3230 - enabling batcherrors parameter with PL/SQL"
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-2040:",
                               self.cursor.executemany, "begin null; end;", 30,
                               batcherrors=True)

    def test_3231_arraydmlrowcounts_with_plsql(self):
        "3230 - enabling arraydmlrowcountsbatcherrors parameter with PL/SQL"
        self.assertRaisesRegex(oracledb.DatabaseError, "^DPY-2040:",
                               self.cursor.executemany, "begin null; end;", 31,
                               arraydmlrowcounts=True)

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
2800 - Module for testing AQ Bulk enqueue/dequeue
"""

import datetime
import threading
import unittest

import oracledb
import test_env

RAW_QUEUE_NAME = "TEST_RAW_QUEUE"
JSON_QUEUE_NAME = "TEST_JSON_QUEUE"
RAW_PAYLOAD_DATA = [
    "The first message",
    "The second message",
    "The third message",
    "The fourth message",
    "The fifth message",
    "The sixth message",
    "The seventh message",
    "The eighth message",
    "The ninth message",
    "The tenth message",
    "The eleventh message",
    "The twelfth and final message"
]

JSON_DATA_PAYLOAD = [
    [
        2.75,
        True,
        'Ocean Beach',
        b'Some bytes',
        {'keyA': 1.0, 'KeyB': 'Melbourne'},
        datetime.datetime(2022, 8, 1, 0, 0)
    ],
    dict(name="John", age=30, city="New York")
]

@unittest.skipIf(test_env.get_is_thin(), "thin mode doesn't support AQ yet")
class TestCase(test_env.BaseTestCase):

    def __deq_in_thread(self, results):
        with test_env.get_connection() as connection:
            queue = connection.queue(RAW_QUEUE_NAME)
            queue.deqoptions.wait = 10
            queue.deqoptions.navigation = oracledb.DEQ_FIRST_MSG
            while len(results) < len(RAW_PAYLOAD_DATA):
                messages = queue.deqmany(5)
                if not messages:
                    break
                for m in messages:
                    results.append(m.payload.decode())
            connection.commit()

    def test_2800_enq_and_deq(self):
        "2800 - test bulk enqueue and dequeue"
        queue = self.get_and_clear_queue(RAW_QUEUE_NAME)
        messages = [self.connection.msgproperties(payload=d) \
                    for d in RAW_PAYLOAD_DATA]
        queue.enqmany(messages)
        messages = queue.deqmany(len(RAW_PAYLOAD_DATA))
        data = [m.payload.decode() for m in messages]
        self.connection.commit()
        self.assertEqual(data, RAW_PAYLOAD_DATA)

    def test_2801_dequeue_empty(self):
        "2801 - test empty bulk dequeue"
        queue = self.get_and_clear_queue(RAW_QUEUE_NAME)
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        messages = queue.deqmany(5)
        self.connection.commit()
        self.assertEqual(messages, [])

    def test_2802_deq_with_wait(self):
        "2802 - test bulk dequeue with wait"
        queue = self.get_and_clear_queue(RAW_QUEUE_NAME)
        results = []
        thread = threading.Thread(target=self.__deq_in_thread, args=(results,))
        thread.start()
        messages = [self.connection.msgproperties(payload=d) \
                    for d in RAW_PAYLOAD_DATA]
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        queue.enqmany(messages)
        thread.join()
        self.assertEqual(results, RAW_PAYLOAD_DATA)

    def test_2803_enq_and_deq_multiple_times(self):
        "2803 - test enqueue and dequeue multiple times"
        queue = self.get_and_clear_queue(RAW_QUEUE_NAME)
        data_to_enqueue = RAW_PAYLOAD_DATA
        for num in (2, 6, 4):
            messages = [self.connection.msgproperties(payload=d) \
                        for d in data_to_enqueue[:num]]
            data_to_enqueue = data_to_enqueue[num:]
            queue.enqmany(messages)
        self.connection.commit()
        all_data = []
        for num in (3, 5, 10):
            messages = queue.deqmany(num)
            all_data.extend(m.payload.decode() for m in messages)
        self.connection.commit()
        self.assertEqual(all_data, RAW_PAYLOAD_DATA)

    def test_2804_enq_and_deq_visibility(self):
        "2804 - test visibility option for enqueue and dequeue"
        queue = self.get_and_clear_queue(RAW_QUEUE_NAME)

        # first test with ENQ_ON_COMMIT (commit required)
        queue.enqoptions.visibility = oracledb.ENQ_ON_COMMIT
        props1 = self.connection.msgproperties(payload="A first message")
        props2 = self.connection.msgproperties(payload="A second message")
        queue.enqmany([props1, props2])
        other_connection = test_env.get_connection()
        other_queue = other_connection.queue(RAW_QUEUE_NAME)
        other_queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        other_queue.deqoptions.visibility = oracledb.DEQ_ON_COMMIT
        messages = other_queue.deqmany(5)
        self.assertEqual(len(messages), 0)
        self.connection.commit()
        messages = other_queue.deqmany(5)
        self.assertEqual(len(messages), 2)
        other_connection.rollback()

        # second test with ENQ_IMMEDIATE (no commit required)
        queue.enqoptions.visibility = oracledb.ENQ_IMMEDIATE
        other_queue.deqoptions.visibility = oracledb.DEQ_IMMEDIATE
        queue.enqmany([props1, props2])
        messages = other_queue.deqmany(5)
        self.assertEqual(len(messages), 4)
        other_connection.rollback()
        messages = other_queue.deqmany(5)
        self.assertEqual(len(messages), 0)

    def test_2805_messages_with_no_payload(self):
        "2805 - test error for messages with no payload"
        queue = self.get_and_clear_queue(RAW_QUEUE_NAME)
        messages = [self.connection.msgproperties() \
                    for d in RAW_PAYLOAD_DATA]
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2000:",
                               queue.enqmany, messages)

    def test_2806_verify_msgid(self):
        "2806 - verify that the msgid property is returned correctly"
        queue = self.get_and_clear_queue(RAW_QUEUE_NAME)
        messages = [self.connection.msgproperties(payload=d) \
                    for d in RAW_PAYLOAD_DATA]
        queue.enqmany(messages)
        self.cursor.execute("select msgid from raw_queue_tab")
        actual_msgids = set(m for m, in self.cursor)
        msgids = set(m.msgid for m in messages)
        self.assertEqual(msgids, actual_msgids)
        messages = queue.deqmany(len(RAW_PAYLOAD_DATA))
        msgids = set(m.msgid for m in messages)
        self.assertEqual(msgids, actual_msgids)

    def test_2807_json_enq_deq(self):
        "4800 - test enqueuing and dequeuing JSON message"
        queue = self.get_and_clear_queue(JSON_QUEUE_NAME, "JSON")
        props = [self.connection.msgproperties(payload=d) \
                 for d in JSON_DATA_PAYLOAD]
        queue.enqmany(props)
        self.connection.commit()
        queue.deqoptions.wait = oracledb.DEQ_NO_WAIT
        messages = queue.deqmany(5)
        actual_data = [m.payload for m in messages]
        self.assertEqual(actual_data, JSON_DATA_PAYLOAD)

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# lob.pyx
#
# Cython file defining the thick Lob implementation class (embedded in
# thick_impl.pyx).
#------------------------------------------------------------------------------

cdef class ThickLobImpl(BaseLobImpl):
    cdef:
        dpiLob *_handle

    @staticmethod
    cdef ThickLobImpl _create(ThickConnImpl conn_impl, DbType dbtype,
                             dpiLob *handle):
        cdef:
            ThickLobImpl impl = ThickLobImpl.__new__(ThickLobImpl)
            int status
        impl.dbtype = dbtype
        if handle == NULL:
            with nogil:
                status = dpiConn_newTempLob(conn_impl._handle, dbtype.num,
                                            &handle)
            if status < 0:
                _raise_from_odpi()
        elif dpiLob_addRef(handle) < 0:
            _raise_from_odpi()
        impl._handle = handle
        return impl

    def close(self):
        """
        Internal method for closing a LOB that was opened earlier.
        """
        cdef int status
        with nogil:
            status = dpiLob_closeResource(self._handle)
        if status < 0:
            _raise_from_odpi()

    def file_exists(self):
        """
        Internal method for returning whether the file referenced by a BFILE
        exists.
        """
        cdef:
            bint exists
            int status
        with nogil:
            status = dpiLob_getFileExists(self._handle, &exists)
        if status < 0:
            _raise_from_odpi()
        return exists

    def free_lob(self):
        """
        Internal method for releasing the handle
        """
        if self._handle != NULL:
            dpiLob_release(self._handle)

    def get_chunk_size(self):
        """
        Internal method for returning the chunk size of the LOB.
        """
        cdef uint32_t chunk_size
        if dpiLob_getChunkSize(self._handle, &chunk_size) < 0:
            _raise_from_odpi()
        return chunk_size

    def get_file_name(self):
        """
        Internal method for returning a 2-tuple constaining the directory alias
        and file name of a BFILE type LOB.
        """
        cdef:
            uint32_t dir_alias_len, file_name_len
            const char *dir_alias
            const char *file_name
            int status
        with nogil:
            status = dpiLob_getDirectoryAndFileName(self._handle, &dir_alias,
                                                    &dir_alias_len,
                                                    &file_name, &file_name_len)
        if status < 0:
            _raise_from_odpi()
        return (dir_alias[:dir_alias_len].decode(),
                file_name[:file_name_len].decode())

    def get_is_open(self):
        """
        Internal method for returning whether the LOB is open or not.
        """
        cdef:
            bint is_open
            int status
        with nogil:
            status = dpiLob_getIsResourceOpen(self._handle, &is_open)
        if status < 0:
            _raise_from_odpi()
        return is_open

    def get_max_amount(self):
        """
        Internal method for returning the maximum amount that can be read.
        """
        return self.get_size()

    def get_size(self):
        """
        Internal method for returning the size of a LOB.
        """
        cdef uint64_t size
        if dpiLob_getSize(self._handle, &size) < 0:
            _raise_from_odpi()
        return size

    def open(self):
        """
        Internal method for opening a LOB.
        """
        cdef int status
        with nogil:
            status = dpiLob_openResource(self._handle)
        if status < 0:
            _raise_from_odpi()

    def read(self, uint64_t offset, uint64_t amount):
        """
        Internal method for reading a portion (or all) of the data in the LOB.
        """
        cdef:
            uint64_t buf_size
            object result
            char *buf
            int status
        if dpiLob_getBufferSize(self._handle, amount, &buf_size) < 0:
            _raise_from_odpi()
        buf = <char*> cpython.PyMem_Malloc(buf_size)
        with nogil:
            status = dpiLob_readBytes(self._handle, offset, amount, buf,
                                      &buf_size)
        try:
            if status < 0:
                _raise_from_odpi()
            result = buf[:buf_size]
            if self.dbtype.num == DPI_ORACLE_TYPE_CLOB \
                    or self.dbtype.num == DPI_ORACLE_TYPE_NCLOB:
                result = result.decode()
            return result
        finally:
            cpython.PyMem_Free(buf)

    def set_file_name(self, str dir_alias, str name):
        """
        Internal method for setting the directory alias and file name
        associated with a BFILE LOB.
        """
        cdef:
            bytes dir_alias_bytes, name_bytes
            uint32_t dir_alias_len, name_len
            const char *dir_alias_ptr
            const char *name_ptr
            int status
        dir_alias_bytes = dir_alias.encode()
        dir_alias_ptr = dir_alias_bytes
        dir_alias_len = <uint32_t> len(dir_alias_bytes)
        name_bytes = name.encode()
        name_ptr = name_bytes
        name_len = <uint32_t> len(name_bytes)
        with nogil:
            status = dpiLob_setDirectoryAndFileName(self._handle,
                                                    dir_alias_ptr,
                                                    dir_alias_len,
                                                    name_ptr, name_len)
        if status < 0:
            _raise_from_odpi()

    def trim(self, uint64_t new_size):
        """
        Internal method for trimming the data in the LOB to the new size
        """
        cdef int status
        with nogil:
            status = dpiLob_trim(self._handle, new_size)
        if status < 0:
            _raise_from_odpi()

    def write(self, object value, uint64_t offset):
        """
        Internal method for writing data to the LOB object.
        """
        cdef:
            StringBuffer buf = StringBuffer()
            int status
        buf.set_value(value)
        with nogil:
            status = dpiLob_writeBytes(self._handle, offset, buf.ptr,
                                       buf.length)
        if status < 0:
            _raise_from_odpi()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2023, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

"""
4000 - Module for testing the cursor executemany() method
"""

import decimal

import oracledb
import test_env

class TestCase(test_env.BaseTestCase):

    def test_4000_executemany_by_name(self):
        "4000 - test executing a statement multiple times (named args)"
        self.cursor.execute("truncate table TestTempTable")
        rows = [{"value": n} for n in range(250)]
        self.cursor.arraysize = 100
        statement = "insert into TestTempTable (IntCol) values (:value)"
        self.cursor.executemany(statement, rows)
        self.connection.commit()
        self.cursor.execute("select count(*) from TestTempTable")
        count, = self.cursor.fetchone()
        self.assertEqual(count, len(rows))

    def test_4001_executemany_by_position(self):
        "4001 - test executing a statement multiple times (positional args)"
        self.cursor.execute("truncate table TestTempTable")
        rows = [[n] for n in range(230)]
        self.cursor.arraysize = 100
        statement = "insert into TestTempTable (IntCol) values (:1)"
        self.cursor.executemany(statement, rows)
        self.connection.commit()
        self.cursor.execute("select count(*) from TestTempTable")
        count, = self.cursor.fetchone()
        self.assertEqual(count, len(rows))

    def test_4002_executemany_with_prepare(self):
        "4002 - test executing a statement multiple times (with prepare)"
        self.cursor.execute("truncate table TestTempTable")
        rows = [[n] for n in range(225)]
        self.cursor.arraysize = 100
        statement = "insert into TestTempTable (IntCol) values (:1)"
        self.cursor.prepare(statement)
        self.cursor.executemany(None, rows)
        self.connection.commit()
        self.cursor.execute("select count(*) from TestTempTable")
        count, = self.cursor.fetchone()
        self.assertEqual(count, len(rows))

    def test_4003_executemany_with_rebind(self):
        "4003 - test executing a statement multiple times (with rebind)"
        self.cursor.execute("truncate table TestTempTable")
        rows = [[n] for n in range(235)]
        self.cursor.arraysize = 100
        statement = "insert into TestTempTable (IntCol) values (:1)"
        self.cursor.executemany(statement, rows[:50])
        self.cursor.executemany(statement, rows[50:])
        self.connection.commit()
        self.cursor.execute("select count(*) from TestTempTable")
        count, = self.cursor.fetchone()
        self.assertEqual(count, len(rows))

    def test_4004_executemany_with_input_sizes_wrong(self):
        "4004 - test executing multiple times (with input sizes wrong)"
        cursor = self.connection.cursor()
        cursor.setinputsizes(oracledb.NUMBER)
        data = [[decimal.Decimal("25.8")], [decimal.Decimal("30.0")]]
        cursor.executemany("declare t number; begin t := :1; end;", data)

    def test_4005_executemany_with_multiple_batches(self):
        "4005 - test executing multiple times (with multiple batches)"
        self.cursor.execute("truncate table TestTempTable")
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        self.cursor.executemany(sql, [(1, None), (2, None)])
        self.cursor.executemany(sql, [(3, None), (4, "Testing")])

    def test_4006_executemany_numeric(self):
        "4006 - test executemany() with various numeric types"
        self.cursor.execute("truncate table TestTempTable")
        data = [
            (1, 5),
            (2, 7.0),
            (3, 6.5),
            (4, 2 ** 65),
            (5, decimal.Decimal("24.5"))
        ]
        sql = "insert into TestTempTable (IntCol, NumberCol) values (:1, :2)"
        self.cursor.executemany(sql, data)
        self.cursor.execute("""
                select IntCol, NumberCol
                from TestTempTable
                order by IntCol""")
        self.assertEqual(self.cursor.fetchall(), data)

    def test_4007_executemany_with_resize(self):
        "4007 - test executing a statement multiple times (with resize)"
        self.cursor.execute("truncate table TestTempTable")
        rows = [
            (1, "First"),
            (2, "Second"),
            (3, "Third"),
            (4, "Fourth"),
            (5, "Fifth"),
            (6, "Sixth"),
            (7, "Seventh and the longest one")
        ]
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        self.cursor.executemany(sql, rows)
        self.cursor.execute("""
                select IntCol, StringCol1
                from TestTempTable
                order by IntCol""")
        fetched_rows = self.cursor.fetchall()
        self.assertEqual(fetched_rows, rows)

    def test_4008_executemany_with_exception(self):
        "4008 - test executing a statement multiple times (with exception)"
        self.cursor.execute("truncate table TestTempTable")
        rows = [{"value": n} for n in (1, 2, 3, 2, 5)]
        statement = "insert into TestTempTable (IntCol) values (:value)"
        self.assertRaisesRegex(oracledb.DatabaseError, "^ORA-00001:",
                               self.cursor.executemany, statement, rows)
        self.assertEqual(self.cursor.rowcount, 3)

    def test_4009_executemany_with_invalid_parameters(self):
        "4009 - test calling executemany() with invalid parameters"
        sql = "insert into TestTempTable (IntCol, StringCol1) values (:1, :2)"
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2004:",
                               self.cursor.executemany, sql,
                               "These are not valid parameters")

    def test_4010_executemany_no_parameters(self):
        "4010 - test calling executemany() without any bind parameters"
        num_rows = 5
        self.cursor.execute("truncate table TestTempTable")
        self.cursor.executemany("""
                declare
                    t_Id          number;
                begin
                    select nvl(count(*), 0) + 1 into t_Id
                    from TestTempTable;

                    insert into TestTempTable (IntCol, StringCol1)
                    values (t_Id, 'Test String ' || t_Id);
                end;""", num_rows)
        self.assertEqual(self.cursor.rowcount, 0)
        self.cursor.execute("select count(*) from TestTempTable")
        count, = self.cursor.fetchone()
        self.assertEqual(count, num_rows)

    def test_4011_executemany_bound_earlier(self):
        "4011 - test calling executemany() with binds performed earlier"
        num_rows = 9
        self.cursor.execute("truncate table TestTempTable")
        var = self.cursor.var(int, arraysize=num_rows)
        self.cursor.setinputsizes(var)
        self.cursor.executemany("""
                declare
                    t_Id          number;
                begin
                    select nvl(count(*), 0) + 1 into t_Id
                    from TestTempTable;

                    insert into TestTempTable (IntCol, StringCol1)
                    values (t_Id, 'Test String ' || t_Id);

                    select sum(IntCol) into :1
                    from TestTempTable;
                end;""", num_rows)
        self.assertEqual(self.cursor.rowcount, 0)
        expected_data = [1, 3, 6, 10, 15, 21, 28, 36, 45]
        self.assertEqual(var.values, expected_data)

    def test_4012_executemany_with_plsql_binds(self):
        "4012 - test executing plsql statements multiple times (with binds)"
        var = self.cursor.var(int, arraysize=5)
        self.cursor.setinputsizes(var)
        data = [[25], [30], [None], [35], [None]]
        exepected_data = [25, 30, None, 35, None]
        self.cursor.executemany("declare t number; begin t := :1; end;", data)
        self.assertEqual(var.values, exepected_data)

    def test_4013_executemany_with_incorrect_params(self):
        "4013 - test executemany with incorrect parameters"
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2004:",
                               self.cursor.executemany, "select :1 from dual",
                               [1])

    def test_4014_executemany_with_mixed_binds_pos_first(self):
        "4014 - test executemany with mixed binds (pos first)"
        rows = [["test"], {"value": 1}]
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2006:",
                               self.cursor.executemany, "select :1 from dual",
                               rows)

    def test_4015_executemany_with_mixed_binds_name_first(self):
        "4015 - test executemany with mixed binds (name first)"
        rows = [{"value": 1}, ["test"]]
        self.assertRaisesRegex(oracledb.ProgrammingError, "^DPY-2006:",
                               self.cursor.executemany,
                               "select :value from dual", rows)

    def test_4016_executemany_plsql_dml_returning(self):
        "4016 - test executemany() with a pl/sql statement with dml returning"
        num_rows = 5
        self.cursor.execute("truncate table TestTempTable")
        out_var = self.cursor.var(oracledb.NUMBER, arraysize=5)
        self.cursor.setinputsizes(out_var)
        self.cursor.executemany("""
                declare
                    t_Id          number;
                begin
                    select nvl(count(*), 0) + 1 into t_Id
                    from TestTempTable;

                    insert into TestTempTable (IntCol, StringCol1)
                    values (t_Id, 'Test String ' || t_Id) returning
                    IntCol into :out_bind;
                end;""", num_rows)
        self.assertEqual(out_var.values, [1, 2, 3, 4, 5])

    def test_4017_executemany_pl_sql_with_in_and_out_binds(self):
        "4017 - test executemany() with pl/sql in binds and out binds"
        self.cursor.execute("truncate table TestTempTable")
        int_values = [5, 8, 17, 24, 6]
        str_values = ["Test 5", "Test 8", "Test 17", "Test 24", "Test 6"]
        out_bind = self.cursor.var(oracledb.NUMBER, arraysize=5)
        self.cursor.setinputsizes(None, None, out_bind)
        data = list(zip(int_values, str_values))
        self.cursor.executemany("""
            begin
                insert into TestTempTable (IntCol, StringCol1)
                values (:int_val, :str_val)
                returning IntCol into :out_bind;
            end;""", data)
        self.assertEqual(out_bind.values, [5, 8, 17, 24, 6])

    def test_4018_executemany_pl_sql_out_bind(self):
        "4018 - test executemany() with pl/sql outbinds"
        self.cursor.execute("truncate table TestTempTable")
        out_bind = self.cursor.var(oracledb.NUMBER, arraysize=5)
        self.cursor.setinputsizes(out_bind)
        self.cursor.executemany("""
            begin
                :out_var := 5;
            end;""", 5)
        self.assertEqual(out_bind.values, [5, 5, 5, 5, 5])

    def test_4019_re_executemany_pl_sql_with_in_and_out_binds(self):
        "4019 - test re-executemany() with pl/sql in binds and out binds"
        int_values = [5, 8, 17, 24, 6]
        str_values = ["Test 5", "Test 8", "Test 17", "Test 24", "Test 6"]
        data = list(zip(int_values, str_values))
        for i in range(2):
            self.cursor.execute("truncate table TestTempTable")
            out_bind = self.cursor.var(oracledb.NUMBER, arraysize=5)
            self.cursor.setinputsizes(None, None, out_bind)
            self.cursor.executemany("""
                begin
                    insert into TestTempTable (IntCol, StringCol1)
                    values (:int_val, :str_val)
                    returning IntCol into :out_bind;
                end;""", data)
            self.assertEqual(out_bind.values, [5, 8, 17, 24, 6])

    def test_4020_executemany_with_plsql_single_row(self):
        "4020 - test PL/SQL statement with single row bind"
        value = 4020
        var = self.cursor.var(int)
        data = [[var, value]]
        self.cursor.executemany("begin :1 := :2; end;", data)
        self.assertEqual(var.values, [value])

    def test_4021_defer_type_assignment(self):
        "4021 - test deferral of type assignment"
        self.cursor.execute("truncate table TestTempTable")
        data = [(1, None), (2, 25)]
        self.cursor.executemany("""
                insert into TestTempTable
                (IntCol, NumberCol)
                values (:1, :2)""", data)
        self.connection.commit()
        self.cursor.execute("""
                select IntCol, NumberCol
                from TestTempTable
                order by IntCol""")
        fetched_data = self.cursor.fetchall()
        self.assertEqual(data, fetched_data)

    def test_4022_plsql_large_number_of_binds(self):
        "4022 - test PL/SQL with a lerge number of binds"
        parts = []
        bind_names = []
        all_bind_values = []
        out_binds = []
        for i in range(5):
            all_bind_values.append([])
        for i in range(350):
            n = len(parts) + 1
            bind_names.extend([f"v_out_{n}_0", f"a_{n}", f"b_{n}", f"c_{n}"])
            parts.append(f":v_out{n} := :a_{n} + :b_{n} + :c_{n};")
            out_binds.append(self.cursor.var(int,
                                             arraysize=len(all_bind_values)))
            for j, bind_values in enumerate(all_bind_values):
                bind_values.extend([out_binds[-1], n * 1 + j, n * 2 + j,
                                    n * 3 + j])
        lf = "\n"
        sql = f"begin{lf}{lf.join(parts)}{lf}end;"
        self.cursor.executemany(sql, all_bind_values)
        init_val = 6
        for var in out_binds:
            expected = [init_val, init_val + 3, init_val + 6, init_val + 9,
                        init_val + 12]
            self.assertEqual(var.values, expected)
            init_val += 6

if __name__ == "__main__":
    test_env.run_test_cases()



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2023, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# utils.pyx
#
# Cython file defining utility classes and methods (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

cdef bint DEBUG_PACKETS = ("PYO_DEBUG_PACKETS" in os.environ)

class ConnectConstants:

    def __init__(self):
        self.program_name = sys.executable
        self.machine_name = socket.gethostname()
        self.pid = str(os.getpid())
        try:
            self.user_name = getpass.getuser()
        except:
            self.user_name = ""
        self.terminal_name = "unknown"
        self.sanitized_program_name = self._sanitize(self.program_name)
        self.sanitized_machine_name = self._sanitize(self.machine_name)
        self.sanitized_user_name = self._sanitize(self.user_name)
        pattern = r"(?P<major_num>\d+)\.(?P<minor_num>\d+)\.(?P<patch_num>\d+)"
        match_dict = re.match(pattern, VERSION)
        major_num = int(match_dict["major_num"])
        minor_num = int(match_dict["minor_num"])
        patch_num = int(match_dict["patch_num"])
        self.full_version_num = \
                major_num << 24 | minor_num << 20 | patch_num << 12

    def _sanitize(self, value):
        """
        Sanitize the value by replacing all the "(", ")" and "=" characters
        with "?". These are invalid characters within connect data.
        """
        return value.replace("(", "?").replace(")", "?").replace("=", "?")

_connect_constants = ConnectConstants()

cdef int _convert_base64(char_type *buf, long value, int size, int offset):
    """
    Converts every 6 bits into a character, from left to right. This is
    similar to ordinary base64 encoding with a few differences and written
    here for performance.
    """
    cdef int i
    for i in range(size):
        buf[offset + size - i - 1] = TNS_BASE64_ALPHABET_ARRAY[value & 0x3f]
        value = value >> 6
    return offset + size


cdef object _encode_rowid(Rowid *rowid):
    """
    Converts the rowid structure into an encoded string, if the rowid structure
    contains valid data; otherwise, it returns None.
    """
    cdef:
        char_type buf[TNS_MAX_ROWID_LENGTH]
        int offset = 0
    if rowid.rba != 0 or rowid.partition_id != 0 or rowid.block_num != 0 \
            or rowid.slot_num != 0:
        offset = _convert_base64(buf, rowid.rba, 6, offset)
        offset = _convert_base64(buf, rowid.partition_id, 3, offset)
        offset = _convert_base64(buf, rowid.block_num, 6, offset)
        offset = _convert_base64(buf, rowid.slot_num, 3, offset)
        return buf[:TNS_MAX_ROWID_LENGTH].decode()


cdef str _get_connect_data(Description description):
    """
    Return the connect data required by the listener in order to connect.
    """
    constants = _connect_constants
    cid = f"(PROGRAM={constants.sanitized_program_name})" + \
          f"(HOST={constants.sanitized_machine_name})" + \
          f"(USER={constants.sanitized_user_name})"
    return description.build_connect_string(cid)


def _print_packet(operation, socket_fileno, data):
    """
    Print the packet content in a format suitable for debugging.
    """
    offset = 0
    hex_data = memoryview(data).hex().upper()
    current_date = datetime.datetime.now().isoformat(sep=" ",
                                                     timespec="milliseconds")
    output_lines = [f"{current_date} [socket: {socket_fileno}] {operation}"]
    while hex_data:
        line_hex_data = hex_data[:16]
        hex_data = hex_data[16:]
        hex_dump_values = []
        printable_values = []
        for i in range(0, len(line_hex_data), 2):
            hex_byte = line_hex_data[i:i + 2]
            hex_dump_values.append(hex_byte)
            byte_val = ord(bytes.fromhex(hex_byte))
            char_byte = chr(byte_val)
            if char_byte.isprintable() and byte_val < 128 \
                    and char_byte != " ":
                printable_values.append(char_byte)
            else:
                printable_values.append(".")
        while len(hex_dump_values) < 8:
            hex_dump_values.append("  ")
            printable_values.append(" ")
        output_lines.append(f'{offset:04} : {" ".join(hex_dump_values)} ' + \
                            f'|{"".join(printable_values)}|')
        offset += 8
    output_lines.append("")
    print("\n".join(output_lines))


def init_thin_impl(package):
    """
    Initializes globals after the package has been completely initialized. This
    is to avoid circular imports and eliminate the need for global lookups.
    """
    global PY_TYPE_DB_OBJECT, PY_TYPE_LOB
    PY_TYPE_DB_OBJECT = package.DbObject
    PY_TYPE_LOB = package.LOB



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# cursor.pyx
#
# Cython file defining the base Cursor implementation class (embedded in
# base_impl.pyx).
#------------------------------------------------------------------------------

cdef class BaseCursorImpl:

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef int _bind_values(self,
                          object cursor,
                          object type_handler,
                          object params,
                          uint32_t num_rows,
                          uint32_t row_num,
                          bint defer_type_assignment) except -1:
        """
        Internal method used for binding values.
        """
        if self.bind_vars is None:
            self.bind_vars = []
        if isinstance(params, dict):
            if self.bind_style is None:
                self.bind_style = dict
                self.bind_vars_by_name = {}
            elif self.bind_style is not dict:
                errors._raise_err(errors.ERR_MIXED_POSITIONAL_AND_NAMED_BINDS)
            self._bind_values_by_name(cursor, type_handler, <dict> params,
                                      num_rows, row_num, defer_type_assignment)
        elif isinstance(params, (list, tuple)):
            if self.bind_style is None:
                self.bind_style = list
            elif self.bind_style is not list:
                errors._raise_err(errors.ERR_MIXED_POSITIONAL_AND_NAMED_BINDS)
            self._bind_values_by_position(cursor, type_handler, <dict> params,
                                          num_rows, row_num,
                                          defer_type_assignment)
        else:
            errors._raise_err(errors.ERR_WRONG_EXECUTEMANY_PARAMETERS_TYPE)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef int _bind_values_by_name(self,
                                  object cursor,
                                  object type_handler,
                                  dict params,
                                  uint32_t num_rows,
                                  uint32_t row_num,
                                  bint defer_type_assignment) except -1:
        """
        Internal method used for binding values by name.
        """
        cdef:
            BindVar bind_var
            object conn
            ssize_t pos
        conn = cursor.connection
        for name, value in params.items():
            bind_var = <BindVar> self.bind_vars_by_name.get(name)
            if bind_var is None:
                pos = len(self.bind_vars_by_name)
                if pos < len(self.bind_vars):
                    bind_var = <BindVar> self.bind_vars[pos]
                else:
                    bind_var = BindVar.__new__(BindVar)
                    self.bind_vars.append(bind_var)
                bind_var.name = name
                self.bind_vars_by_name[name] = bind_var
            bind_var._set_by_value(conn, self, cursor, value, type_handler,
                                   row_num, num_rows, defer_type_assignment)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef int _bind_values_by_position(self,
                                      object cursor,
                                      object type_handler,
                                      object params,
                                      uint32_t num_rows,
                                      uint32_t row_num,
                                      bint defer_type_assignment) except -1:
        """
        Internal method used for binding values by position.
        """
        cdef:
            BindVar bind_var
            object conn
            ssize_t i
        conn = cursor.connection
        for i, value in enumerate(params):
            if i < len(self.bind_vars):
                bind_var = <BindVar> self.bind_vars[i]
            else:
                bind_var = BindVar.__new__(BindVar)
                bind_var.pos = i + 1
                self.bind_vars.append(bind_var)
            bind_var._set_by_value(conn, self, cursor, value, type_handler,
                                   row_num, num_rows, defer_type_assignment)

    cdef int _close(self, bint in_del) except -1:
        """
        Internal method for closing the cursor.
        """
        raise NotImplementedError()

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef int _create_fetch_var(self, object conn, object cursor,
                               object type_handler, ssize_t pos,
                               FetchInfo fetch_info) except -1:
        """
        Create the fetch variable for the given position and fetch information.
        The output type handler is consulted, if present, to make any necessary
        adjustments.
        """
        cdef:
            BaseVarImpl var_impl
            uint32_t db_type_num
            object var

        # if an output type handler is specified, call it; the output type
        # handler should return a variable or None; the value None implies that
        # the default processing should take place just as if no output type
        # handler was defined
        if type_handler is not None:
            var = type_handler(cursor, fetch_info._name, fetch_info._dbtype,
                               fetch_info._size, fetch_info._precision,
                               fetch_info._scale)
            if var is not None:
                self._verify_var(var)
                var_impl = var._impl
                var_impl._fetch_info = fetch_info
                self.fetch_vars[pos] = var
                self.fetch_var_impls[pos] = var_impl
                return 0

        # otherwise, create a new variable using the provided fetch information
        var_impl = self._create_var_impl(conn)
        var_impl.num_elements = self._fetch_array_size
        var_impl.dbtype = fetch_info._dbtype
        var_impl.objtype = fetch_info._objtype
        var_impl.name = fetch_info._name
        var_impl.size = fetch_info._size
        var_impl.precision = fetch_info._precision
        var_impl.scale = fetch_info._scale
        var_impl.nulls_allowed = fetch_info._nulls_allowed
        var_impl._fetch_info = fetch_info

        # adjust the variable based on the defaults specified by the user, if
        # applicable
        db_type_num = var_impl.dbtype.num
        if db_type_num == DB_TYPE_NUM_NUMBER:
            if defaults.fetch_decimals:
                var_impl._preferred_num_type = NUM_TYPE_DECIMAL
            elif var_impl.scale == 0 \
                    or (var_impl.scale == -127 and var_impl.precision == 0):
                var_impl._preferred_num_type = NUM_TYPE_INT
        elif not defaults.fetch_lobs:
            if db_type_num == DB_TYPE_NUM_BLOB:
                var_impl.dbtype = DB_TYPE_LONG_RAW
            elif db_type_num == DB_TYPE_NUM_CLOB:
                var_impl.dbtype = DB_TYPE_LONG
            elif db_type_num == DB_TYPE_NUM_NCLOB:
                var_impl.dbtype = DB_TYPE_LONG_NVARCHAR

        # finalize variable and store in arrays
        var_impl._finalize_init()
        self.fetch_var_impls[pos] = var_impl
        self.fetch_vars[pos] = PY_TYPE_VAR._from_impl(var_impl)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef object _create_row(self):
        """
        Internal method for creating a row from the fetched data.
        """
        cdef:
            Py_ssize_t i, num_vars
            BaseVarImpl var_impl
            object row, value
        num_vars = cpython.PyList_GET_SIZE(self.fetch_var_impls)
        row = cpython.PyTuple_New(num_vars)
        for i in range(num_vars):
            var_impl = self.fetch_var_impls[i]
            value = var_impl._get_scalar_value(self._buffer_index)
            cpython.Py_INCREF(value)
            cpython.PyTuple_SET_ITEM(row, i, value)
        if self.rowfactory is not None:
            row = self.rowfactory(*row)
        self._buffer_index += 1
        self._buffer_rowcount -= 1
        self.rowcount += 1
        return row

    cdef BaseVarImpl _create_var_impl(self, object conn):
        """
        Internal method for creating a variable.
        """
        raise NotImplementedError()

    cdef int _fetch_rows(self, object cursor) except -1:
        """
        Internal method used for fetching rows from a cursor.
        """
        raise NotImplementedError()

    cdef BaseConnImpl _get_conn_impl(self):
        """
        Internal method used to return the connection implementation associated
        with the cursor implementation.
        """
        raise NotImplementedError()

    cdef object _get_input_type_handler(self):
        """
        Return the input type handler to use for the cursor. If one is not
        directly defined on the cursor then the one defined on the connection
        is used instead.
        """
        cdef BaseConnImpl conn_impl
        if self.inputtypehandler is not None:
            return self.inputtypehandler
        conn_impl = self._get_conn_impl()
        return conn_impl.inputtypehandler

    @utils.CheckImpls("getting a cursor OCI attribute")
    def _get_oci_attr(self, uint32_t attr_num, uint32_t attr_type):
        pass

    cdef object _get_output_type_handler(self):
        """
        Return the output type handler to use for the cursor. If one is not
        directly defined on the cursor then the one defined on the connection
        is used instead.
        """
        cdef BaseConnImpl conn_impl
        if self.outputtypehandler is not None:
            return self.outputtypehandler
        conn_impl = self._get_conn_impl()
        return conn_impl.outputtypehandler

    cdef int _init_fetch_vars(self, uint32_t num_columns) except -1:
        """
        Initializes the fetch variable lists in preparation for creating the
        fetch variables used in fetching rows from the database.
        """
        self.fetch_vars = [None] * num_columns
        self.fetch_var_impls = [None] * num_columns

    cdef bint _is_plsql(self):
        """
        Internal method that indicates whether the currently prepared statement
        is a PL/SQL statement or not.
        """
        raise NotImplementedError()

    cdef int _perform_binds(self, object conn, uint32_t num_execs) except -1:
        """
        Perform all binds on the cursor.
        """
        cdef:
            BindVar bind_var
            ssize_t i
        for i, bind_var in enumerate(self.bind_vars):
            bind_var.var_impl._bind(conn, self, num_execs, bind_var.name,
                                    bind_var.pos)

    cdef int _reset_bind_vars(self, uint32_t num_rows) except -1:
        """
        Reset all of the existing bind variables. If any bind variables don't
        have enough space to store the number of rows specified, expand and
        then reinitialize that bind variable.
        """
        cdef:
            BindVar bind_var
            ssize_t i
        if self.bind_vars is not None:
            for i in range(len(self.bind_vars)):
                bind_var = <BindVar> self.bind_vars[i]
                if bind_var.var_impl is not None:
                    bind_var.var_impl._on_reset_bind(num_rows)
                bind_var.has_value = False

    @utils.CheckImpls("setting a cursor OCI attribute")
    def _set_oci_attr(self, uint32_t attr_num, uint32_t attr_type,
                      object value):
        pass

    cdef int _verify_var(self, object var) except -1:
        """
        Internal method used for verifying if an outputtypehandler returns a
        valid var object.
        """
        if not isinstance(var, PY_TYPE_VAR):
            errors._raise_err(errors.ERR_EXPECTING_VAR)
        if self.arraysize > var.num_elements:
            errors._raise_err(errors.ERR_INCORRECT_VAR_ARRAYSIZE,
                              var_arraysize=var.num_elements,
                              required_arraysize=self.arraysize)

    def bind_many(self, object cursor, list parameters):
        """
        Internal method used for binding multiple rows of data.
        """
        cdef:
            bint defer_type_assignment
            ssize_t i, num_rows
            object params_row
        type_handler = self._get_input_type_handler()
        num_rows = len(parameters)
        self._reset_bind_vars(num_rows)
        for i, params_row in enumerate(parameters):
            defer_type_assignment = (i < num_rows - 1)
            self._bind_values(cursor, type_handler, params_row, num_rows, i,
                              defer_type_assignment)

    def bind_one(self, cursor, parameters):
        """
        Internal method used for binding a single row of data.
        """
        cdef:
            bint defer_type_assignment = False
            uint32_t row_num = 0, num_rows = 1
            ssize_t num_bind_vars, pos
            object name, value
            BindVar bind_var
            dict dict_params
        type_handler = self._get_input_type_handler()
        self._reset_bind_vars(num_rows)
        self._bind_values(cursor, type_handler, parameters, num_rows, row_num,
                          defer_type_assignment)

    def close(self, bint in_del=False):
        """
        Closes the cursor and makes it unusable for further operations.
        """
        self.bind_vars = None
        self.bind_vars_by_name = None
        self.bind_style = None
        self.fetch_vars = None
        self._close(in_del)

    def create_var(self, object conn, object typ, uint32_t size=0,
                   uint32_t num_elements=1, object inconverter=None,
                   object outconverter=None, str encoding_errors=None,
                   bint bypass_decode=False, bint is_array=False):
        cdef BaseVarImpl var_impl
        var_impl = self._create_var_impl(conn)
        var_impl._set_type_info_from_type(typ)
        var_impl.size = size
        var_impl.num_elements = num_elements
        var_impl.inconverter = inconverter
        var_impl.outconverter = outconverter
        var_impl.bypass_decode = bypass_decode
        var_impl.is_array = is_array
        var_impl._finalize_init()
        return PY_TYPE_VAR._from_impl(var_impl)

    @utils.CheckImpls("executing a statement")
    def execute(self, cursor):
        pass

    @utils.CheckImpls("executing a statement in batch")
    def executemany(self, cursor, num_execs, batcherrors, arraydmlrowcounts):
        pass

    def fetch_next_row(self, cursor):
        """
        Internal method used for fetching the next row from a cursor.
        """
        if self._buffer_rowcount == 0 and self._more_rows_to_fetch:
            self._fetch_rows(cursor)
        if self._buffer_rowcount > 0:
            return self._create_row()

    @utils.CheckImpls("getting a list of array DML row counts")
    def get_array_dml_row_counts(self):
        pass

    @utils.CheckImpls("getting a list of batch errors")
    def get_batch_errors(self):
        pass

    @utils.CheckImpls("getting a list of bind variable names")
    def get_bind_names(self):
        pass

    def get_bind_vars(self):
        """
        Return a list (when binding by position) or a dictionary (when binding
        by name) of the bind variables associated with the cursor.
        """
        cdef:
            BindVar bind_var
            ssize_t i
        if self.bind_vars is None:
            return []
        for bind_var in self.bind_vars:
            if bind_var.var is None and bind_var.var_impl is not None:
                bind_var.var = PY_TYPE_VAR._from_impl(bind_var.var_impl)
        if self.bind_style is list:
            return [bind_var.var for bind_var in self.bind_vars]
        return dict([(bind_var.name, bind_var.var) \
                for bind_var in self.bind_vars])

    def get_description(self):
        """
        Internal method for populating the cursor description.
        """
        return [v._impl.get_description() for v in self.fetch_vars]

    @utils.CheckImpls("getting implicit results from PL/SQL")
    def get_implicit_results(self, connection):
        pass

    @utils.CheckImpls("getting the rowid of the row last modified")
    def get_lastrowid(self):
        pass

    @utils.CheckImpls("determining if the cursor last executed a query")
    def is_query(self, cursor):
        pass

    @utils.CheckImpls("parsing a statement without executing it")
    def parse(self, cursor):
        pass

    @utils.CheckImpls("preparing a statement")
    def prepare(self, str statement, str tag, bint cache_statement):
        pass

    @utils.CheckImpls("scrolling a scrollable cursor")
    def scroll(self, conn, value, mode):
        """
        Scrolls a scrollable cursor.
        """
        pass

    def setinputsizes(self, object conn, tuple args, dict kwargs):
        """
        Sets type information for bind variables in advance of executing a
        statement (and binding values).
        """
        cdef:
            object name, value
            BindVar bind_var
            ssize_t pos
        self.bind_vars = []
        if kwargs:
            self.bind_style = dict
            self.bind_vars_by_name = {}
            for name, value in kwargs.items():
                bind_var = BindVar.__new__(BindVar)
                self.bind_vars.append(bind_var)
                self.bind_vars_by_name[name] = bind_var
                bind_var._set_by_type(conn, self, value)
                bind_var.name = name
        else:
            self.bind_style = list
            for pos, value in enumerate(args):
                bind_var = BindVar.__new__(BindVar)
                self.bind_vars.append(bind_var)
                bind_var._set_by_type(conn, self, value)
                bind_var.pos = pos + 1



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2022, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# queue.pyx
#
# Cython file defining the thick Queue implementation class (embedded in
# thick_impl.pyx).
#------------------------------------------------------------------------------

cdef class ThickQueueImpl(BaseQueueImpl):
    cdef:
        dpiQueue* _handle
        ThickConnImpl _conn_impl

    def __dealloc__(self):
        if self._handle != NULL:
            dpiQueue_release(self._handle)

    def deq_many(self, uint32_t max_num_messages):
        """
        Internal method for dequeuing multiple message from a queue.
        """
        cdef:
            uint32_t num_props = max_num_messages, processed_props = 0, i
            dpiMsgProps** handles
            ThickMsgPropsImpl props
            ssize_t num_bytes
            list result
            int status
        result = []
        num_bytes = num_props * sizeof(dpiMsgProps*)
        handles = <dpiMsgProps**> cpython.PyMem_Malloc(num_bytes)
        try:
            with nogil:
                status = dpiQueue_deqMany(self._handle, &num_props, handles)
            if status < 0:
                _raise_from_odpi()
            for i in range(num_props):
                props = ThickMsgPropsImpl.__new__(ThickMsgPropsImpl)
                props._handle = handles[i]
                processed_props += 1
                props._initialize(self)
                result.append(props)
        finally:
            for i in range(processed_props, num_props):
                dpiMsgProps_release(handles[i])
            cpython.PyMem_Free(handles)
        return result

    def deq_one(self):
        """
        Internal method for dequeuing a single message from a queue.
        """
        cdef:
            ThickMsgPropsImpl props
            int status
        props = ThickMsgPropsImpl.__new__(ThickMsgPropsImpl)
        with nogil:
            status = dpiQueue_deqOne(self._handle, &props._handle)
        if status < 0:
            _raise_from_odpi()
        if props._handle != NULL:
            props._initialize(self)
            return props

    def enq_many(self, list props_impls):
        """
        Internal method for enqueuing many messages into a queue.
        """
        cdef:
            uint32_t i, num_props = 0
            ThickMsgPropsImpl props
            dpiMsgProps **handles
            ssize_t num_bytes
            int status
        num_props = <uint32_t> len(props_impls)
        num_bytes = num_props * sizeof(dpiMsgProps*)
        handles = <dpiMsgProps**> cpython.PyMem_Malloc(num_bytes)
        for i, props in enumerate(props_impls):
            handles[i] = props._handle
        with nogil:
            status = dpiQueue_enqMany(self._handle, num_props, handles)
        cpython.PyMem_Free(handles)
        if status < 0:
            _raise_from_odpi()

    def enq_one(self, ThickMsgPropsImpl props_impl):
        """
        Internal method for enqueuing a single message into a queue.
        """
        cdef int status
        with nogil:
            status = dpiQueue_enqOne(self._handle, props_impl._handle)
        if status < 0:
            _raise_from_odpi()

    def initialize(self, ThickConnImpl conn_impl, str name,
                   ThickDbObjectTypeImpl payload_type, bint is_json):
        """
        Internal method for initializing the queue.
        """
        cdef:
            ThickDeqOptionsImpl deq_options_impl
            ThickEnqOptionsImpl enq_options_impl
            dpiObjectType* type_handle = NULL
            StringBuffer buf = StringBuffer()
        self._conn_impl = conn_impl
        self.is_json = is_json
        buf.set_value(name)
        if is_json:
            if dpiConn_newJsonQueue(conn_impl._handle, buf.ptr, buf.length,
                                    &self._handle) < 0:
                _raise_from_odpi()
        else:
            if payload_type is not None:
                type_handle = payload_type._handle
            if dpiConn_newQueue(conn_impl._handle, buf.ptr, buf.length,
                                type_handle, &self._handle) < 0:
                _raise_from_odpi()
        deq_options_impl = ThickDeqOptionsImpl.__new__(ThickDeqOptionsImpl)
        if dpiQueue_getDeqOptions(self._handle, &deq_options_impl._handle) < 0:
            _raise_from_odpi()
        self.deq_options_impl = deq_options_impl
        enq_options_impl = ThickEnqOptionsImpl.__new__(ThickEnqOptionsImpl)
        if dpiQueue_getEnqOptions(self._handle, &enq_options_impl._handle) < 0:
            _raise_from_odpi()
        self.enq_options_impl = enq_options_impl
        self.payload_type = payload_type
        self.name = name


cdef class ThickDeqOptionsImpl(BaseDeqOptionsImpl):
    cdef:
        dpiDeqOptions* _handle

    def __dealloc__(self):
        if self._handle != NULL:
            dpiDeqOptions_release(self._handle)

    def get_condition(self):
        """
        Internal method for getting the condition.
        """
        cdef:
            uint32_t value_length
            const char *value
        if dpiDeqOptions_getCondition(self._handle, &value, &value_length) < 0:
            _raise_from_odpi()
        if value != NULL:
            return value[:value_length].decode()

    def get_consumer_name(self):
        """
        Internal method for getting the consumer name.
        """
        cdef:
            uint32_t value_length
            const char *value
        if dpiDeqOptions_getConsumerName(self._handle, &value,
                                         &value_length) < 0:
            _raise_from_odpi()
        if value != NULL:
            return value[:value_length].decode()

    def get_correlation(self):
        """
        Internal method for getting the correlation.
        """
        cdef:
            uint32_t value_length
            const char *value
        if dpiDeqOptions_getCorrelation(self._handle, &value,
                                        &value_length) < 0:
            _raise_from_odpi()
        if value != NULL:
            return value[:value_length].decode()

    def get_message_id(self):
        """
        Internal method for getting the message id.
        """
        cdef:
            uint32_t value_length
            const char *value
        if dpiDeqOptions_getMsgId(self._handle, &value, &value_length) < 0:
            _raise_from_odpi()
        if value != NULL:
            return value[:value_length]

    def get_mode(self):
        """
        Internal method for getting the mode.
        """
        cdef uint32_t value
        if dpiDeqOptions_getMode(self._handle, &value) < 0:
            _raise_from_odpi()
        return value

    def get_navigation(self):
        """
        Internal method for getting the navigation.
        """
        cdef uint32_t value
        if dpiDeqOptions_getNavigation(self._handle, &value) < 0:
            _raise_from_odpi()
        return value

    def get_transformation(self):
        """
        Internal method for getting the transformation.
        """
        cdef:
            uint32_t value_length
            const char *value
        if dpiDeqOptions_getTransformation(self._handle, &value,
                                           &value_length) < 0:
            _raise_from_odpi()
        if value != NULL:
            return value[:value_length].decode()

    def get_visibility(self):
        """
        Internal method for getting the visibility.
        """
        cdef uint32_t value
        if dpiDeqOptions_getVisibility(self._handle, &value) < 0:
            _raise_from_odpi()
        return value

    def get_wait(self):
        """
        Internal method for getting the wait.
        """
        cdef uint32_t value
        if dpiDeqOptions_getWait(self._handle, &value) < 0:
            _raise_from_odpi()
        return value

    def set_condition(self, str value):
        """
        Internal method for setting the condition.
        """
        cdef StringBuffer buf = StringBuffer()
        buf.set_value(value)
        if dpiDeqOptions_setCondition(self._handle, buf.ptr, buf.length) < 0:
            _raise_from_odpi()

    def set_consumer_name(self, str value):
        """
        Internal method for setting the consumer name.
        """
        cdef StringBuffer buf = StringBuffer()
        buf.set_value(value)
        if dpiDeqOptions_setConsumerName(self._handle, buf.ptr,
                                         buf.length) < 0:
            _raise_from_odpi()

    def set_correlation(self, str value):
        """
        Internal method for setting the correlation.
        """
        cdef StringBuffer buf = StringBuffer()
        buf.set_value(value)
        if dpiDeqOptions_setCorrelation(self._handle, buf.ptr, buf.length) < 0:
            _raise_from_odpi()

    def set_delivery_mode(self, uint16_t value):
        """
        Internal method for setting the delivery mode.
        """
        if dpiDeqOptions_setDeliveryMode(self._handle, value) < 0:
            _raise_from_odpi()

    def set_mode(self, uint32_t value):
        """
        Internal method for setting the mode.
        """
        if dpiDeqOptions_setMode(self._handle, value) < 0:
            _raise_from_odpi()

    def set_message_id(self, bytes value):
        """
        Internal method for setting the message id.
        """
        cdef StringBuffer buf = StringBuffer()
        buf.set_value(value)
        if dpiDeqOptions_setMsgId(self._handle, buf.ptr, buf.length) < 0:
            _raise_from_odpi()

    def set_navigation(self, uint32_t value):
        """
        Internal method for setting the navigation.
        """
        if dpiDeqOptions_setNavigation(self._handle, value) < 0:
            _raise_from_odpi()

    def set_transformation(self, str value):
        """
        Internal method for setting the transformation.
        """
        cdef StringBuffer buf = StringBuffer()
        buf.set_value(value)
        if dpiDeqOptions_setTransformation(self._handle, buf.ptr,
                                           buf.length) < 0:
            _raise_from_odpi()

    def set_visibility(self, uint32_t value):
        """
        Internal method for setting the visibility.
        """
        if dpiDeqOptions_setVisibility(self._handle, value) < 0:
            _raise_from_odpi()

    def set_wait(self, uint32_t value):
        """
        Internal method for setting the wait.
        """
        if dpiDeqOptions_setWait(self._handle, value) < 0:
            _raise_from_odpi()


cdef class ThickEnqOptionsImpl(BaseEnqOptionsImpl):
    cdef:
        dpiEnqOptions* _handle

    def __dealloc__(self):
        if self._handle != NULL:
            dpiEnqOptions_release(self._handle)

    def get_transformation(self):
        """
        Internal method for getting the transformation.
        """
        cdef:
            uint32_t value_length
            const char *value
        if dpiEnqOptions_getTransformation(self._handle, &value,
                                           &value_length) < 0:
            _raise_from_odpi()
        if value != NULL:
            return value[:value_length].decode()

    def get_visibility(self):
        """
        Internal method for getting the visibility.
        """
        cdef uint32_t value
        if dpiEnqOptions_getVisibility(self._handle, &value) < 0:
            _raise_from_odpi()
        return value

    def set_delivery_mode(self, uint16_t value):
        """
        Internal method for setting the delivery mode.
        """
        if dpiEnqOptions_setDeliveryMode(self._handle, value) < 0:
            _raise_from_odpi()

    def set_transformation(self, str value):
        """
        Internal method for setting the transformation.
        """
        cdef StringBuffer buf = StringBuffer()
        buf.set_value(value)
        if dpiEnqOptions_setTransformation(self._handle, buf.ptr,
                                           buf.length) < 0:
            _raise_from_odpi()

    def set_visibility(self, uint32_t value):
        """
        Internal method for setting the visibility.
        """
        if dpiEnqOptions_setVisibility(self._handle, value) < 0:
            _raise_from_odpi()


cdef class ThickMsgPropsImpl(BaseMsgPropsImpl):
    cdef:
        dpiMsgProps* _handle
        ThickConnImpl _conn_impl

    def __dealloc__(self):
        if self._handle != NULL:
            dpiMsgProps_release(self._handle)

    cdef int _initialize(self, ThickQueueImpl queue_impl) except -1:
        cdef:
            dpiObject *payload_obj_handle
            ThickDbObjectImpl obj_impl
            const char *payload_ptr
            uint32_t payload_len
            dpiJsonNode *node
            dpiJson *json

        self._conn_impl = queue_impl._conn_impl
        if queue_impl.is_json:
            if dpiMsgProps_getPayloadJson(self._handle, &json) < 0:
                _raise_from_odpi()
            if dpiJson_getValue(json, DPI_JSON_OPT_NUMBER_AS_STRING,
                                &node) < 0:
                _raise_from_odpi()
            self.payload = _convert_from_json_node(node)
        else:
            if dpiMsgProps_getPayload(self._handle, &payload_obj_handle,
                                      &payload_ptr, &payload_len) < 0:
                _raise_from_odpi()
            if payload_obj_handle != NULL:
                obj_impl = ThickDbObjectImpl.__new__(ThickDbObjectImpl)
                obj_impl.type = queue_impl.payload_type
                if dpiObject_addRef(payload_obj_handle) < 0:
                    _raise_from_odpi()
                obj_impl._handle = payload_obj_handle
                self.payload = PY_TYPE_DB_OBJECT._from_impl(obj_impl)
            else:
                self.payload = payload_ptr[:payload_len]

    def get_num_attempts(self):
        """
        Internal method for getting the number of attempts made.
        """
        cdef int32_t value
        if dpiMsgProps_getNumAttempts(self._handle, &value) < 0:
            _raise_from_odpi()
        return value

    def get_correlation(self):
        """
        Internal method for getting the correlation.
        """
        cdef:
            uint32_t value_length
            const char *value
        if dpiMsgProps_getCorrelation(self._handle, &value, &value_length) < 0:
            _raise_from_odpi()
        if value != NULL:
            return value[:value_length].decode()

    def get_delay(self):
        """
        Internal method for getting the delay.
        """
        cdef int32_t value
        if dpiMsgProps_getDelay(self._handle, &value) < 0:
            _raise_from_odpi()
        return value

    def get_delivery_mode(self):
        """
        Internal method for getting the delivery mode.
        """
        cdef uint16_t value
        if dpiMsgProps_getDeliveryMode(self._handle, &value) < 0:
            _raise_from_odpi()
        return value

    def get_enq_time(self):
        """
        Internal method for getting the enqueue time.
        """
        cdef dpiTimestamp timestamp
        if dpiMsgProps_getEnqTime(self._handle, &timestamp) < 0:
            _raise_from_odpi()
        return cydatetime.datetime_new(timestamp.year, timestamp.month,
                                       timestamp.day, timestamp.hour,
                                       timestamp.minute, timestamp.second,
                                       timestamp.fsecond // 1000, None)

    def get_exception_queue(self):
        """
        Internal method for getting the exception queue.
        """
        cdef:
            uint32_t value_length
            const char *value
        if dpiMsgProps_getExceptionQ(self._handle, &value, &value_length) < 0:
            _raise_from_odpi()
        if value != NULL:
            return value[:value_length].decode()

    def get_expiration(self):
        """
        Internal method for getting the message expiration.
        """
        cdef int32_t value
        if dpiMsgProps_getExpiration(self._handle, &value) < 0:
            _raise_from_odpi()
        return value

    def get_message_id(self):
        """
        Internal method for getting the message id.
        """
        cdef:
            uint32_t value_length
            const char *value
        if dpiMsgProps_getMsgId(self._handle, &value, &value_length) < 0:
            _raise_from_odpi()
        if value != NULL:
            return value[:value_length]

    def get_priority(self):
        """
        Internal method for getting the priority.
        """
        cdef int32_t value
        if dpiMsgProps_getPriority(self._handle, &value) < 0:
            _raise_from_odpi()
        return value

    def get_state(self):
        """
        Internal method for getting the message state.
        """
        cdef uint32_t value
        if dpiMsgProps_getState(self._handle, &value) < 0:
            _raise_from_odpi()
        return value

    def set_correlation(self, str value):
        """
        Internal method for setting the correlation.
        """
        cdef StringBuffer buf = StringBuffer()
        buf.set_value(value)
        if dpiMsgProps_setCorrelation(self._handle, buf.ptr, buf.length) < 0:
            _raise_from_odpi()

    def set_delay(self, int32_t value):
        """
        Internal method for setting the correlation.
        """
        if dpiMsgProps_setDelay(self._handle, value) < 0:
            _raise_from_odpi()

    def set_exception_queue(self, str value):
        """
        Internal method for setting the exception queue.
        """
        cdef StringBuffer buf = StringBuffer()
        buf.set_value(value)
        if dpiMsgProps_setExceptionQ(self._handle, buf.ptr, buf.length) < 0:
            _raise_from_odpi()

    def set_expiration(self, int32_t value):
        """
        Internal method for setting the message expiration.
        """
        if dpiMsgProps_setExpiration(self._handle, value) < 0:
            _raise_from_odpi()

    def set_payload_bytes(self, bytes value):
        """
        Internal method for setting the payload from bytes.
        """
        cdef StringBuffer buf = StringBuffer()
        buf.set_value(value)
        if dpiMsgProps_setPayloadBytes(self._handle, buf.ptr, buf.length) < 0:
            _raise_from_odpi()

    def set_payload_object(self, ThickDbObjectImpl obj_impl):
        """
        Internal method for setting the payload from an object.
        """
        if dpiMsgProps_setPayloadObject(self._handle, obj_impl._handle) < 0:
            _raise_from_odpi()

    def set_payload_json(self, object json_val):
        """
        Internal method for setting the payload from a JSON object
        """
        cdef:
            JsonBuffer json_buf = JsonBuffer()
            dpiJson *json

        json_buf.from_object(json_val)
        if dpiConn_newJson(self._conn_impl._handle, &json) < 0:
            _raise_from_odpi()
        if dpiJson_setValue(json, &json_buf._top_node) < 0:
            _raise_from_odpi()
        if dpiMsgProps_setPayloadJson(self._handle, json) < 0:
            _raise_from_odpi()

    def set_priority(self, int32_t value):
        """
        Internal method for setting the priority.
        """
        if dpiMsgProps_setPriority(self._handle, value) < 0:
            _raise_from_odpi()

    def set_recipients(self, list value):
        """
        Internal method for setting the recipients list.
        """
        cdef:
            dpiMsgRecipient *recipients
            uint32_t num_recipients
            ssize_t num_bytes
            list buffers = []
            StringBuffer buf

        num_recipients = <uint32_t> len(value)
        num_bytes = num_recipients * sizeof(dpiMsgRecipient)
        recipients = <dpiMsgRecipient*> cpython.PyMem_Malloc(num_bytes)
        try:
            for i in range(num_recipients):
                buf = StringBuffer()
                buf.set_value(value[i])
                buffers.append(buf)
                recipients[i].name = buf.ptr
                recipients[i].nameLength = buf.length
            if dpiMsgProps_setRecipients(self._handle, recipients,
                                         num_recipients) < 0:
                _raise_from_odpi()
        finally:
            cpython.PyMem_Free(recipients)
